import asyncio
import getpass
from datetime import datetime

import click
import pyhanko.sign
from asn1crypto import cms, pem
from pyhanko.cli._trust import (
    _get_key_usage_settings,
    _prepare_vc,
    build_vc_kwargs,
    trust_options,
)
from pyhanko.cli.commands.signing import signing
from pyhanko.cli.runtime import pyhanko_exception_manager
from pyhanko.cli.utils import logger
from pyhanko.pdf_utils import crypt
from pyhanko.pdf_utils.misc import isoparse
from pyhanko.pdf_utils.reader import PdfFileReader
from pyhanko.sign import validation
from pyhanko.sign.validation import RevocationInfoValidationType
from pyhanko.sign.validation.errors import SignatureValidationError

from pyhanko_certvalidator import ValidationContext

__all__ = ['validate_signatures']


def _signature_status(
    ltv_profile,
    vc_kwargs,
    force_revinfo,
    key_usage_settings,
    embedded_sig,
    skip_diff=False,
):
    if ltv_profile is None:
        vc = ValidationContext(**vc_kwargs)
        status = pyhanko.sign.validation.validate_pdf_signature(
            embedded_sig,
            key_usage_settings=key_usage_settings,
            signer_validation_context=vc,
            skip_diff=skip_diff,
        )
    else:
        status = validation.validate_pdf_ltv_signature(
            embedded_sig,
            ltv_profile,
            key_usage_settings=key_usage_settings,
            force_revinfo=force_revinfo,
            validation_context_kwargs=vc_kwargs,
            skip_diff=skip_diff,
        )
    return status


def _validate_detached(
    infile, sig_infile, validation_context, key_usage_settings
):
    sig_bytes = sig_infile.read()
    try:
        if pem.detect(sig_bytes):
            _, _, sig_bytes = pem.unarmor(sig_bytes)
        content_info = cms.ContentInfo.load(sig_bytes)
        if content_info['content_type'].native != 'signed_data':
            raise click.ClickException("CMS content type is not signedData")
    except ValueError as e:
        raise click.ClickException("Could not parse CMS object") from e

    validation_coro = validation.async_validate_detached_cms(
        infile,
        signed_data=content_info['content'],
        signer_validation_context=validation_context,
        key_usage_settings=key_usage_settings,
    )
    return asyncio.run(validation_coro)


def _signature_status_str(status_callback, pretty_print, executive_summary):
    try:
        status = status_callback()
        if executive_summary and not pretty_print:
            return (
                'VALID' if status.bottom_line else 'INVALID',
                status.bottom_line,
            )
        elif pretty_print:
            return status.pretty_print_details(), status.bottom_line
        else:
            return status.summary(), status.bottom_line
    except validation.ValidationInfoReadingError as e:
        msg = (
            'An error occurred while parsing the revocation information '
            'for this signature: ' + str(e)
        )
        logger.error(msg)
        if pretty_print:
            return msg, False
        else:
            return 'REVINFO_FAILURE', False
    except SignatureValidationError as e:
        msg = 'An error occurred while validating this signature: ' + str(e)
        logger.error(msg, exc_info=e)
        if pretty_print:
            return msg, False
        else:
            return 'INVALID', False


def _attempt_iso_dt_parse(dt_str) -> datetime:
    try:
        dt = isoparse(dt_str)
    except ValueError:
        raise click.ClickException(f"datetime {dt_str!r} could not be parsed")
    return dt


# TODO add an option to do LTV, but guess the profile


@trust_options
@signing.command(name='validate', help='validate signatures')
@click.argument('infile', type=click.File('rb'))
@click.option(
    '--executive-summary',
    help='only print final judgment on signature validity',
    type=bool,
    is_flag=True,
    default=False,
    show_default=True,
)
@click.option(
    '--pretty-print',
    help='render a prettier summary for the signatures in the file',
    type=bool,
    is_flag=True,
    default=False,
    show_default=True,
)
@click.option(
    '--ltv-profile',
    help='LTV signature validation profile',
    type=click.Choice(RevocationInfoValidationType.as_tuple()),
    required=False,
)
@click.option(
    '--force-revinfo',
    help='Fail trust validation if a certificate has no known CRL '
    'or OCSP endpoints.',
    type=bool,
    is_flag=True,
    default=False,
    show_default=True,
)
@click.option(
    '--soft-revocation-check',
    help='Do not fail validation on revocation checking failures '
    '(only applied to on-line revocation checks)',
    type=bool,
    is_flag=True,
    default=False,
    show_default=True,
)
@click.option(
    '--no-revocation-check',
    help='Do not attempt to check revocation status '
    '(meaningless for LTV validation)',
    type=bool,
    is_flag=True,
    default=False,
    show_default=True,
)
@click.option(
    '--retroactive-revinfo',
    help='Treat revocation info as retroactively valid '
    '(i.e. ignore thisUpdate timestamp)',
    type=bool,
    is_flag=True,
    default=False,
    show_default=True,
)
@click.option(
    '--validation-time',
    help=(
        'Override the validation time (ISO 8601 date). '
        'The special value \'claimed\' causes the validation time '
        'claimed by the signer to be used. Revocation checking '
        'will be disabled. Option ignored in LTV mode.'
    ),
    type=str,
    required=False,
)
@click.option(
    '--password',
    required=False,
    type=str,
    help='password to access the file (can also be read from stdin)',
)
@click.option(
    '--no-diff-analysis',
    default=False,
    type=bool,
    is_flag=True,
    help='disable incremental update analysis',
)
@click.option(
    '--detached',
    type=click.File('rb'),
    help=(
        'Read signature CMS object from the indicated file; '
        'this can be used to verify signatures on non-PDF files'
    ),
)
@click.option(
    '--no-strict-syntax',
    help='Attempt to ignore syntactical problems in the input file '
    'and enable signature validation in hybrid-reference files.'
    '(warning: this may affect validation results in unexpected '
    'ways.)',
    type=bool,
    is_flag=True,
    default=False,
    show_default=True,
)
@click.pass_context
def validate_signatures(
    ctx: click.Context,
    infile,
    executive_summary,
    pretty_print,
    validation_context,
    trust,
    trust_replace,
    other_certs,
    ltv_profile,
    force_revinfo,
    soft_revocation_check,
    no_revocation_check,
    password,
    retroactive_revinfo,
    detached,
    no_diff_analysis,
    validation_time,
    no_strict_syntax,
):
    no_revocation_check |= validation_time is not None

    if no_revocation_check:
        soft_revocation_check = True

    if pretty_print and executive_summary:
        raise click.ClickException(
            "--pretty-print is incompatible with --executive-summary."
        )

    if ltv_profile is not None:
        if validation_time is not None:
            raise click.ClickException(
                "--validation-time is not compatible with --ltv-profile"
            )
        ltv_profile = RevocationInfoValidationType(ltv_profile)

    vc_kwargs = build_vc_kwargs(
        ctx.obj.config,
        validation_context,
        trust,
        trust_replace,
        other_certs,
        retroactive_revinfo,
        allow_fetching=False if no_revocation_check else None,
    )

    use_claimed_validation_time = False
    if validation_time == 'claimed':
        use_claimed_validation_time = True
    elif validation_time is not None:
        vc_kwargs['moment'] = _attempt_iso_dt_parse(validation_time)

    key_usage_settings = _get_key_usage_settings(ctx, validation_context)
    vc_kwargs = _prepare_vc(
        vc_kwargs,
        soft_revocation_check=soft_revocation_check,
        force_revinfo=force_revinfo,
    )
    with pyhanko_exception_manager():
        if detached is not None:
            (status_str, signature_ok) = _signature_status_str(
                status_callback=lambda: _validate_detached(
                    infile,
                    detached,
                    ValidationContext(**vc_kwargs),
                    key_usage_settings,
                ),
                pretty_print=pretty_print,
                executive_summary=executive_summary,
            )
            if signature_ok:
                click.echo(status_str)
            else:
                raise click.ClickException(status_str)
            return

        if no_strict_syntax:
            logger.info(
                "Strict PDF syntax is disabled; this could impact validation "
                "results. Use caution."
            )
            r = PdfFileReader(infile, strict=False)
        else:
            r = PdfFileReader(infile)
        sh = r.security_handler
        if isinstance(sh, crypt.StandardSecurityHandler):
            if password is None:
                password = getpass.getpass(prompt='File password: ')
            auth_result = r.decrypt(password)
            if auth_result.status == crypt.AuthStatus.FAILED:
                raise click.ClickException("Password didn't match.")
        elif sh is not None:
            raise click.ClickException(
                "The CLI supports only password-based encryption when "
                "validating (for now)"
            )

        all_signatures_ok = True
        for ix, embedded_sig in enumerate(r.embedded_regular_signatures):
            fingerprint: str = embedded_sig.signer_cert.sha256.hex()
            if use_claimed_validation_time:
                vc_kwargs['moment'] = embedded_sig.self_reported_timestamp
            (status_str, signature_ok) = _signature_status_str(
                status_callback=lambda: _signature_status(
                    ltv_profile=ltv_profile,
                    force_revinfo=force_revinfo,
                    vc_kwargs=vc_kwargs,
                    key_usage_settings=key_usage_settings,
                    embedded_sig=embedded_sig,
                    skip_diff=no_diff_analysis,
                ),
                pretty_print=pretty_print,
                executive_summary=executive_summary,
            )
            name = embedded_sig.field_name

            if pretty_print:
                header = f'Field {ix + 1}: {name}'
                line = '=' * len(header)
                click.echo(line)
                click.echo(header)
                click.echo(line)
                click.echo('\n\n' + status_str)
            else:
                click.echo('%s:%s:%s' % (name, fingerprint, status_str))
            all_signatures_ok &= signature_ok

        if not all_signatures_ok:
            raise click.ClickException("Validation failed")



import hashlib
from dataclasses import dataclass
from datetime import timedelta
from typing import Iterable, Optional, Union

import pyhanko.config.pkcs11
import pytest
import yaml
from asn1crypto import x509
from pyhanko import stamp
from pyhanko.cli import config
from pyhanko.cli.commands.signing.pkcs11_cli import ModuleConfigWrapper
from pyhanko.cli.commands.signing.simple import KeyFileConfigWrapper
from pyhanko.config.api import ConfigurableMixin
from pyhanko.config.errors import ConfigurationError
from pyhanko.config.logging import DEFAULT_ROOT_LOGGER_LEVEL, StdLogOutput
from pyhanko.config.pkcs11 import TokenCriteria
from pyhanko.config.trust import (
    DEFAULT_TIME_TOLERANCE,
    init_validation_context_kwargs,
)
from pyhanko.pdf_utils import layout
from pyhanko.pdf_utils.content import ImportedPdfPage
from pyhanko.pdf_utils.images import PdfImage
from pyhanko.sign.signers.pdf_cms import (
    signer_from_p12_config,
    signer_from_pemder_config,
)
from pyhanko.stamp import QRStampStyle, TextStampStyle

from .samples import CRYPTO_DATA_DIR, TEST_DIR, TESTING_CA_DIR


def _parse_cli_config(config_string):
    return config.parse_cli_config(config_string).config


@pytest.mark.parametrize('trust_replace', [True, False])
def test_read_vc_kwargs(trust_replace):
    config_string = f"""
    validation-contexts:
        default:
            trust: '{TESTING_CA_DIR}/root/root.cert.pem'
            trust-replace: {'true' if trust_replace else 'false'}
            other-certs: '{TESTING_CA_DIR}/ca-chain.cert.pem'
    """
    cli_config: config.CLIConfig = _parse_cli_config(config_string)
    vc_kwargs = cli_config.get_validation_context(as_dict=True)
    assert len(vc_kwargs['other_certs']) == 2
    if trust_replace:
        assert 'extra_trust_roots' not in vc_kwargs
        assert len(vc_kwargs['trust_roots']) == 1
    else:
        assert 'trust_roots' not in vc_kwargs
        assert len(vc_kwargs['extra_trust_roots']) == 1

    ku = cli_config.get_signer_key_usages()
    assert ku.key_usage is None
    assert ku.extd_key_usage is None

    with pytest.raises(ConfigurationError):
        cli_config.get_validation_context('theresnosuchvc')


def test_read_qr_config():
    from pyhanko.pdf_utils.font import SimpleFontEngineFactory
    from pyhanko.pdf_utils.font.opentype import GlyphAccumulatorFactory

    from .test_text import NOTO_SERIF_JP

    config_string = f"""
    stamp-styles:
        default:
            text-box-style:
                font: {NOTO_SERIF_JP}
            type: qr
            background: __stamp__
            qr-position: right
            inner-content-layout:
                y-align: bottom
                x-align: mid
                margins:
                    left: 10
                    right: 10
        alternative1:
            text-box-style:
                font: {NOTO_SERIF_JP}
            background: {TEST_DIR}/data/img/stamp-indexed.png
            type: qr
        alternative2:
            type: qr
            background: {TEST_DIR}/data/pdf/pdf-background-test.pdf
        alternative3:
            type: text
        wrong-position:
            type: qr
            qr-position: bleh
    """
    cli_config: config.CLIConfig = _parse_cli_config(config_string)
    default_qr_style = cli_config.get_stamp_style()
    assert isinstance(default_qr_style, QRStampStyle)
    assert default_qr_style.background is stamp.STAMP_ART_CONTENT
    assert isinstance(
        default_qr_style.text_box_style.font, GlyphAccumulatorFactory
    )
    assert default_qr_style.qr_position == stamp.QRPosition.RIGHT_OF_TEXT

    expected_layout = layout.SimpleBoxLayoutRule(
        x_align=layout.AxisAlignment.ALIGN_MID,
        y_align=layout.AxisAlignment.ALIGN_MIN,
        margins=layout.Margins(left=10, right=10),
    )
    assert default_qr_style.inner_content_layout == expected_layout

    alternative1 = cli_config.get_stamp_style('alternative1')
    assert isinstance(alternative1, QRStampStyle)
    assert isinstance(alternative1.background, PdfImage)
    assert isinstance(alternative1.text_box_style.font, GlyphAccumulatorFactory)
    assert alternative1.qr_position == stamp.QRPosition.LEFT_OF_TEXT

    alternative2 = cli_config.get_stamp_style('alternative2')
    assert isinstance(alternative2, QRStampStyle)
    assert isinstance(alternative2.background, ImportedPdfPage)
    assert isinstance(alternative2.text_box_style.font, SimpleFontEngineFactory)

    alternative3 = cli_config.get_stamp_style('alternative3')
    assert isinstance(alternative3, TextStampStyle)
    assert alternative3.background is None
    assert isinstance(alternative3.text_box_style.font, SimpleFontEngineFactory)

    with pytest.raises(ConfigurationError, match='not a valid QR position'):
        cli_config.get_stamp_style('wrong-position')

    with pytest.raises(ConfigurationError):
        cli_config.get_stamp_style('theresnosuchstyle')


def test_read_bad_config():
    config_string = f"""
    stamp-styles:
        default:
            type: qr
            blah: blah
    """
    cli_config: config.CLIConfig = _parse_cli_config(config_string)
    with pytest.raises(ConfigurationError):
        cli_config.get_stamp_style()


def test_read_bad_background_config():
    config_string = f"""
    stamp-styles:
        default:
            type: text
            background: 1234
    """
    cli_config: config.CLIConfig = _parse_cli_config(config_string)
    with pytest.raises(ConfigurationError, match='must be a string'):
        cli_config.get_stamp_style()


@pytest.mark.parametrize("bad_type", ['5', '[1,2,3]'])
def test_read_bad_config2(bad_type):
    config_string = f"""
    stamp-styles:
        default: {bad_type}
    """
    cli_config: config.CLIConfig = _parse_cli_config(config_string)
    with pytest.raises(ConfigurationError):
        cli_config.get_stamp_style()


def test_empty_config():
    cli_config: config.CLIConfig = _parse_cli_config("")
    vc_kwargs = cli_config.get_validation_context(as_dict=True)
    assert 'extra_trust_roots' not in vc_kwargs
    assert 'trust_roots' not in vc_kwargs
    assert 'other_certs' not in vc_kwargs


def test_read_logging_config():
    config_string = """
    logging:
        root-level: DEBUG
        root-output: stdout
        by-module:
            example.test1:
                level: 50
                output: test.log
            example.test2:
                level: DEBUG
            example.test3:
                level: 10
                output: stderr
    """
    cli_config: config.CLIRootConfig = config.parse_cli_config(config_string)

    assert cli_config.log_config[None].output == StdLogOutput.STDOUT
    assert cli_config.log_config[None].level == 'DEBUG'

    assert cli_config.log_config['example.test1'].level == 50
    assert cli_config.log_config['example.test1'].output == 'test.log'
    assert cli_config.log_config['example.test2'].level == 'DEBUG'
    assert cli_config.log_config['example.test2'].output == StdLogOutput.STDERR
    assert cli_config.log_config['example.test3'].level == 10
    assert cli_config.log_config['example.test3'].output == StdLogOutput.STDERR


def test_read_logging_config_defaults():
    cli_config: config.CLIRootConfig = config.parse_cli_config(
        """
        logging:
            root-level: DEBUG
    """
    )

    assert cli_config.log_config[None].output == StdLogOutput.STDERR
    assert cli_config.log_config[None].level == 'DEBUG'
    assert list(cli_config.log_config.keys()) == [None]

    cli_config: config.CLIRootConfig = config.parse_cli_config(
        """
        logging:
            root-output: 'test.log'
    """
    )

    assert cli_config.log_config[None].output == 'test.log'
    assert cli_config.log_config[None].level == DEFAULT_ROOT_LOGGER_LEVEL
    assert list(cli_config.log_config.keys()) == [None]

    cli_config: config.CLIRootConfig = config.parse_cli_config("")
    assert cli_config.log_config[None].output == StdLogOutput.STDERR
    assert cli_config.log_config[None].level == DEFAULT_ROOT_LOGGER_LEVEL
    assert list(cli_config.log_config.keys()) == [None]


WRONG_CONFIGS = [
    # a bunch of type errors
    "logging: 5",
    """
    logging:
        by-module: 1
    """,
    """
    logging:
        root-output: [1, 2]
    """,
    """
    logging:
        root-level: [2, 3]
    """,
    """
    logging:
        by-module:
            test.example:
                level: 10
                output: 5
    """,
    """
    logging:
        by-module:
            0:
                level: 10
                
    """,
    # level is required for non-root logging specs
    """
    logging:
        by-module:
            test.example:
                output: 'abc.log'
    """,
]


@pytest.mark.parametrize('config_str', WRONG_CONFIGS)
def test_read_logging_config_errors(config_str):
    with pytest.raises(ConfigurationError):
        _parse_cli_config(config_str)


@pytest.mark.parametrize(
    'key_usage_str, key_usages',
    [
        ('non_repudiation', {'non_repudiation'}),
        (
            '[non_repudiation, digital_signature]',
            {'non_repudiation', 'digital_signature'},
        ),
        ('[]', set()),
    ],
)
def test_read_key_usage(key_usage_str, key_usages):
    config_string = f"""
    validation-contexts:
        default:
            trust: '{TESTING_CA_DIR}/root/root.cert.pem'
            signer-key-usage: {key_usage_str}
    """
    cli_config: config.CLIConfig = _parse_cli_config(config_string)
    key_usage_settings = cli_config.get_signer_key_usages()
    assert key_usage_settings.key_usage == key_usages
    assert key_usage_settings.extd_key_usage is None


@pytest.mark.parametrize(
    'key_usage_str, key_usages',
    [
        ('piv_content_signing', {'piv_content_signing'}),
        (
            '[piv_content_signing, code_signing]',
            {'piv_content_signing', 'code_signing'},
        ),
        ('[]', set()),
        (
            '[2.16.840.1.101.3.6.7, code_signing]',
            {'piv_content_signing', 'code_signing'},
        ),
        ('[2.16.840.1.101.3.6.7, "2.999"]', {'piv_content_signing', '2.999'}),
        ('2.16.840.1.101.3.6.7', {'piv_content_signing'}),
        ('"2.999"', {'2.999'}),
    ],
)
def test_read_extd_key_usage(key_usage_str, key_usages):
    config_string = f"""
    validation-contexts:
        default:
            trust: '{TESTING_CA_DIR}/root/root.cert.pem'
            signer-extd-key-usage: {key_usage_str}
    """
    cli_config: config.CLIConfig = _parse_cli_config(config_string)
    key_usage_settings = cli_config.get_signer_key_usages()
    assert key_usage_settings.key_usage is None
    assert key_usage_settings.extd_key_usage == key_usages
    assert not key_usage_settings.match_all_key_usages


def test_read_key_usage_policy_1():
    config_string = f"""
    validation-contexts:
        default:
            trust: '{TESTING_CA_DIR}/root/root.cert.pem'
            signer-key-usage-policy:
                key-usage: [digital_signature, non_repudiation]
                match-all-key-usages: true
    """
    cli_config: config.CLIConfig = _parse_cli_config(config_string)
    key_usage_settings = cli_config.get_signer_key_usages()
    assert key_usage_settings.key_usage == {
        'digital_signature',
        'non_repudiation',
    }
    assert key_usage_settings.key_usage_forbidden is None
    assert key_usage_settings.extd_key_usage is None
    assert key_usage_settings.match_all_key_usages


def test_read_key_usage_policy_2():
    config_string = f"""
    validation-contexts:
        default:
            trust: '{TESTING_CA_DIR}/root/root.cert.pem'
            signer-key-usage-policy:
                key-usage: [digital_signature, non_repudiation]
                extd-key-usage: '2.999'
                explicit-extd-key-usage-required: true
    """
    cli_config: config.CLIConfig = _parse_cli_config(config_string)
    key_usage_settings = cli_config.get_signer_key_usages()
    assert key_usage_settings.key_usage == {
        'digital_signature',
        'non_repudiation',
    }
    assert key_usage_settings.extd_key_usage == {'2.999'}
    assert key_usage_settings.explicit_extd_key_usage_required


def test_read_key_usage_policy_3():
    config_string = f"""
    validation-contexts:
        default:
            trust: '{TESTING_CA_DIR}/root/root.cert.pem'
            signer-key-usage-policy:
                key-usage: [digital_signature, non_repudiation]
                key-usage-forbidden: data_encipherment
    """
    cli_config: config.CLIConfig = _parse_cli_config(config_string)
    key_usage_settings = cli_config.get_signer_key_usages()
    assert key_usage_settings.key_usage == {
        'digital_signature',
        'non_repudiation',
    }
    assert key_usage_settings.key_usage_forbidden == {'data_encipherment'}
    assert key_usage_settings.extd_key_usage is None
    assert not key_usage_settings.match_all_key_usages


@pytest.mark.parametrize(
    'key_usage_str',
    [
        '0',
        '["non_repudiation", 2]',
        "[1, 2, 3]",
        "abcdef",
        '["no_such_key_usage"]',
    ],
)
def test_extd_key_usage_errors(key_usage_str):
    config_string = f"""
    validation-contexts:
        default:
            trust: '{TESTING_CA_DIR}/root/root.cert.pem'
            signer-extd-key-usage: {key_usage_str}
    """
    cli_config: config.CLIConfig = _parse_cli_config(config_string)
    with pytest.raises(ConfigurationError):
        cli_config.get_signer_key_usages()


@pytest.mark.parametrize(
    'key_usage_str',
    [
        '0',
        '["non_repudiation", 2]',
        "[1, 2, 3]",
        "abcdef",
        '["no_such_key_usage"]',
    ],
)
def test_key_usage_errors(key_usage_str):
    config_string = f"""
    validation-contexts:
        default:
            trust: '{TESTING_CA_DIR}/root/root.cert.pem'
            signer-key-usage: {key_usage_str}
    """
    cli_config: config.CLIConfig = _parse_cli_config(config_string)
    with pytest.raises(ConfigurationError):
        cli_config.get_signer_key_usages()


@pytest.mark.parametrize(
    'config_string, result',
    [
        (
            f"""
    time-tolerance: 5
    validation-contexts:
        default:
            trust: '{TESTING_CA_DIR}/root/root.cert.pem'
            other-certs: '{TESTING_CA_DIR}/ca-chain.cert.pem'
    """,
            5,
        ),
        (
            f"""
    time-tolerance: 5
    validation-contexts:
        default:
            time-tolerance: 7
            trust: '{TESTING_CA_DIR}/root/root.cert.pem'
            other-certs: '{TESTING_CA_DIR}/ca-chain.cert.pem'
    """,
            7,
        ),
        (
            f"""
    validation-contexts:
        default:
            time-tolerance: 7
            trust: '{TESTING_CA_DIR}/root/root.cert.pem'
            other-certs: '{TESTING_CA_DIR}/ca-chain.cert.pem'
    """,
            7,
        ),
        (
            f"""
    validation-contexts:
        default:
            trust: '{TESTING_CA_DIR}/root/root.cert.pem'
            other-certs: '{TESTING_CA_DIR}/ca-chain.cert.pem'
    """,
            DEFAULT_TIME_TOLERANCE.seconds,
        ),
    ],
)
def test_read_time_tolerance(config_string, result):
    cli_config: config.CLIConfig = _parse_cli_config(config_string)
    vc_kwargs = cli_config.get_validation_context(as_dict=True)
    assert vc_kwargs['time_tolerance'] == timedelta(seconds=result)


def test_read_time_tolerance_input_issues():
    config_string = f"""
    validation-contexts:
        default:
            time-tolerance: "this makes no sense"
            trust: '{TESTING_CA_DIR}/root/root.cert.pem'
    """
    with pytest.raises(ConfigurationError, match='time-tolerance.*'):
        cli_config: config.CLIConfig = _parse_cli_config(config_string)
        cli_config.get_validation_context(as_dict=True)

    config_string = f"""
    time-tolerance: "this makes no sense"
    validation-contexts:
        default:
            trust: '{TESTING_CA_DIR}/root/root.cert.pem'
    """
    with pytest.raises(ConfigurationError, match='time-tolerance.*'):
        cli_config: config.CLIConfig = _parse_cli_config(config_string)
        cli_config.get_validation_context(as_dict=True)

    vc_kwargs = init_validation_context_kwargs(
        trust=[], trust_replace=False, other_certs=[]
    )
    assert 'retroactive_revinfo' not in vc_kwargs
    assert vc_kwargs['time_tolerance'] == DEFAULT_TIME_TOLERANCE


@pytest.mark.parametrize(
    'config_string, result',
    [
        (
            f"""
    retroactive-revinfo: true
    validation-contexts:
        default:
            trust: '{TESTING_CA_DIR}/root/root.cert.pem'
            other-certs: '{TESTING_CA_DIR}/ca-chain.cert.pem'
    """,
            True,
        ),
        (
            f"""
    retroactive-revinfo: true
    validation-contexts:
        default:
            retroactive-revinfo: false
            trust: '{TESTING_CA_DIR}/root/root.cert.pem'
            other-certs: '{TESTING_CA_DIR}/ca-chain.cert.pem'
    """,
            False,
        ),
        (
            f"""
    retroactive-revinfo: false
    validation-contexts:
        default:
            retroactive-revinfo: true
            trust: '{TESTING_CA_DIR}/root/root.cert.pem'
            other-certs: '{TESTING_CA_DIR}/ca-chain.cert.pem'
    """,
            True,
        ),
        (
            f"""
    validation-contexts:
        default:
            retroactive-revinfo: true
            trust: '{TESTING_CA_DIR}/root/root.cert.pem'
            other-certs: '{TESTING_CA_DIR}/ca-chain.cert.pem'
    """,
            True,
        ),
        (
            f"""
    validation-contexts:
        default:
            retroactive-revinfo: "yes"
            trust: '{TESTING_CA_DIR}/root/root.cert.pem'
            other-certs: '{TESTING_CA_DIR}/ca-chain.cert.pem'
    """,
            True,
        ),
        (
            f"""
    validation-contexts:
        default:
            trust: '{TESTING_CA_DIR}/root/root.cert.pem'
            other-certs: '{TESTING_CA_DIR}/ca-chain.cert.pem'
    """,
            False,
        ),
    ],
)
def test_read_retroactive_revinfo(config_string, result):
    cli_config: config.CLIConfig = _parse_cli_config(config_string)
    vc_kwargs = cli_config.get_validation_context(as_dict=True)
    if result is False:
        assert 'retroactive_revinfo' not in vc_kwargs
    else:
        assert vc_kwargs['retroactive_revinfo']


def test_read_pkcs11_config():
    cli_config = _parse_cli_config(
        f"""
        pkcs11-setups:
            foo:
                module-path: /path/to/libfoo.so
                token-criteria:
                    label: testrsa
                cert-label: signer
                other-certs: '{TESTING_CA_DIR}/ca-chain.cert.pem'
        """
    )
    setup = ModuleConfigWrapper(cli_config).get_pkcs11_config('foo')
    with pytest.raises(ConfigurationError):
        ModuleConfigWrapper(cli_config).get_pkcs11_config('bar')

    assert setup.token_criteria == TokenCriteria('testrsa')
    assert setup.module_path == '/path/to/libfoo.so'
    assert len(setup.other_certs) == 2


def test_read_pkcs11_config_legacy():
    cli_config = _parse_cli_config(
        f"""
        pkcs11-setups:
            foo:
                module-path: /path/to/libfoo.so
                token-label: testrsa
                cert-label: signer
                other-certs: '{TESTING_CA_DIR}/ca-chain.cert.pem'
        """
    )
    with pytest.deprecated_call():
        setup = ModuleConfigWrapper(cli_config).get_pkcs11_config('foo')

    assert setup.token_criteria == TokenCriteria('testrsa')


def test_read_pkcs11_config_legacy_with_extra_criteria():
    cli_config = _parse_cli_config(
        f"""
        pkcs11-setups:
            foo:
                module-path: /path/to/libfoo.so
                token-label: testrsa
                token-criteria:
                    serial: deadbeef
                cert-label: signer
                other-certs: '{TESTING_CA_DIR}/ca-chain.cert.pem'
        """
    )
    with pytest.deprecated_call():
        setup = ModuleConfigWrapper(cli_config).get_pkcs11_config('foo')

    assert setup.token_criteria == TokenCriteria(
        label='testrsa', serial=b'\xde\xad\xbe\xef'
    )


def test_read_pkcs11_config_slot_no():
    cli_config = _parse_cli_config(
        f"""
        pkcs11-setups:
            foo:
                module-path: /path/to/libfoo.so
                slot-no: 0
                cert-label: signer
                other-certs: '{TESTING_CA_DIR}/ca-chain.cert.pem'
        """
    )
    setup = ModuleConfigWrapper(cli_config).get_pkcs11_config('foo')

    assert setup.module_path == '/path/to/libfoo.so'
    assert setup.slot_no == 0
    assert len(setup.other_certs) == 2


def test_read_pkcs11_nothing_to_pull():
    cli_config = _parse_cli_config(
        f"""
        pkcs11-setups:
            foo:
                module-path: /path/to/libfoo.so
                token-criteria:
                    label: testrsa
                cert-label: signer
        """
    )
    setup = ModuleConfigWrapper(cli_config).get_pkcs11_config('foo')
    assert len(setup.other_certs) == 0


def test_read_pkcs11_config_ids():
    cli_config = _parse_cli_config(
        f"""
        pkcs11-setups:
            foo:
                module-path: /path/to/libfoo.so
                slot-no: 0
                cert-id: 5
                key-id: "74657374"
        """
    )
    setup = ModuleConfigWrapper(cli_config).get_pkcs11_config('foo')
    assert setup.cert_id == b'\x05'
    assert setup.key_id == b'test'


def test_read_pkcs11_config_external_cert():
    cli_config = _parse_cli_config(
        f"""
        pkcs11-setups:
            foo:
                module-path: /path/to/libfoo.so
                slot-no: 0
                key-id: 10
                signing-certificate: '{TESTING_CA_DIR}/interm/signer1.cert.pem'
        """
    )
    setup = ModuleConfigWrapper(cli_config).get_pkcs11_config('foo')
    assert setup.cert_id is None
    assert setup.cert_label is None
    assert isinstance(setup.signing_certificate, x509.Certificate)


def test_read_pkcs11_config_bad_criteria_type():
    cli_config = _parse_cli_config(
        f"""
        pkcs11-setups:
            foo:
                module-path: /path/to/libfoo.so
                token-criteria: bleh
                cert-label: signer
                other-certs: '{TESTING_CA_DIR}/ca-chain.cert.pem'
        """
    )
    with pytest.raises(
        ConfigurationError, match="TokenCriteria requires a dictionary"
    ):
        ModuleConfigWrapper(cli_config).get_pkcs11_config('foo')


def test_read_pkcs11_config_bad_serial():
    cli_config = _parse_cli_config(
        f"""
        pkcs11-setups:
            foo:
                module-path: /path/to/libfoo.so
                token-criteria:
                    serial: bazz
                cert-label: signer
                other-certs: '{TESTING_CA_DIR}/ca-chain.cert.pem'
        """
    )
    with pytest.raises(ConfigurationError, match="Failed to parse.*hex"):
        ModuleConfigWrapper(cli_config).get_pkcs11_config('foo')


def test_read_pkcs11_config_no_cert_spec_or_key_spec():
    cli_config = _parse_cli_config(
        f"""
        pkcs11-setups:
            foo:
                module-path: /path/to/libfoo.so
                slot-no: 0
        """
    )
    with pytest.raises(ConfigurationError, match="Either 'key_id'"):
        ModuleConfigWrapper(cli_config).get_pkcs11_config('foo')


def test_read_pkcs11_config_cert_label_from_key_label():
    cli_config = _parse_cli_config(
        f"""
        pkcs11-setups:
            foo:
                module-path: /path/to/libfoo.so
                slot-no: 0
                key-label: signer
        """
    )
    cfg = ModuleConfigWrapper(cli_config).get_pkcs11_config('foo')
    assert (cfg.cert_label, cfg.cert_id) == ('signer', None)


def test_read_pkcs11_config_cert_id_from_key_id():
    cli_config = _parse_cli_config(
        f"""
        pkcs11-setups:
            foo:
                module-path: /path/to/libfoo.so
                slot-no: 0
                key-id: deadbeef
        """
    )
    cfg = ModuleConfigWrapper(cli_config).get_pkcs11_config('foo')
    assert (cfg.cert_label, cfg.cert_id) == (None, b"\xde\xad\xbe\xef")


def test_read_pkcs11_config_key_id_from_cert_id():
    cli_config = _parse_cli_config(
        f"""
        pkcs11-setups:
            foo:
                module-path: /path/to/libfoo.so
                slot-no: 0
                cert-id: "deadbeef"
        """
    )
    cfg = ModuleConfigWrapper(cli_config).get_pkcs11_config('foo')
    assert (cfg.key_label, cfg.key_id) == (None, b"\xde\xad\xbe\xef")


def test_read_pkcs11_config_key_label_from_cert_label():
    cli_config = _parse_cli_config(
        f"""
        pkcs11-setups:
            foo:
                module-path: /path/to/libfoo.so
                slot-no: 0
                cert-label: "signer"
        """
    )
    cfg = ModuleConfigWrapper(cli_config).get_pkcs11_config('foo')
    assert (cfg.key_label, cfg.key_id) == ("signer", None)


def test_read_pkcs11_config_key_label_not_from_cert_label_if_key_id_defined():
    cli_config = _parse_cli_config(
        f"""
        pkcs11-setups:
            foo:
                module-path: /path/to/libfoo.so
                slot-no: 0
                cert-label: "signer"
                key-id: "deadbeef"
        """
    )
    cfg = ModuleConfigWrapper(cli_config).get_pkcs11_config('foo')
    assert (cfg.key_label, cfg.key_id) == (None, b"\xde\xad\xbe\xef")
    assert (cfg.cert_label, cfg.cert_id) == ("signer", None)


@pytest.mark.parametrize(
    'literal,exp_val',
    [
        ('prompt', pyhanko.config.pkcs11.PKCS11PinEntryMode.PROMPT),
        ('skip', pyhanko.config.pkcs11.PKCS11PinEntryMode.SKIP),
        ('defer', pyhanko.config.pkcs11.PKCS11PinEntryMode.DEFER),
        # fallbacks
        ('true', pyhanko.config.pkcs11.PKCS11PinEntryMode.PROMPT),
        ('false', pyhanko.config.pkcs11.PKCS11PinEntryMode.SKIP),
        ('1', pyhanko.config.pkcs11.PKCS11PinEntryMode.PROMPT),
        ('0', pyhanko.config.pkcs11.PKCS11PinEntryMode.SKIP),
    ],
)
def test_read_pkcs11_prompt_pin(literal, exp_val):
    cli_config = _parse_cli_config(
        f"""
        pkcs11-setups:
            foo:
                module-path: /path/to/libfoo.so
                token-criteria:
                    label: testrsa
                cert-label: signer
                prompt-pin: {literal}
        """
    )
    setup = ModuleConfigWrapper(cli_config).get_pkcs11_config('foo')
    assert setup.prompt_pin == exp_val


def test_read_pkcs11_prompt_pin_default():
    cli_config = _parse_cli_config(
        f"""
        pkcs11-setups:
            foo:
                module-path: /path/to/libfoo.so
                token-criteria:
                    label: testrsa
                cert-label: signer
        """
    )
    setup = ModuleConfigWrapper(cli_config).get_pkcs11_config('foo')
    assert setup.prompt_pin == pyhanko.config.pkcs11.PKCS11PinEntryMode.PROMPT


def test_read_pkcs11_prompt_pin_invalid():
    cli_config = _parse_cli_config(
        f"""
        pkcs11-setups:
            foo:
                module-path: /path/to/libfoo.so
                token-criteria:
                    label: testrsa
                cert-label: signer
                prompt-pin: foobar
        """
    )
    with pytest.raises(ConfigurationError, match='Invalid'):
        ModuleConfigWrapper(cli_config).get_pkcs11_config('foo')


def _signer_sanity_check(signer):
    digest = hashlib.sha256(b'Hello world!').digest()
    with pytest.deprecated_call():
        sig = signer.sign(digest, digest_algorithm='sha256')
    from pyhanko.sign.validation.generic_cms import validate_sig_integrity

    intact, valid = validate_sig_integrity(
        sig['content']['signer_infos'][0],
        cert=signer.signing_cert,
        expected_content_type='data',
        actual_digest=digest,
    )
    assert intact and valid


def test_read_pkcs12_config():
    cli_config = _parse_cli_config(
        f"""
        pkcs12-setups:
            foo:
                pfx-file: '{TESTING_CA_DIR}/interm/signer1.pfx'
                other-certs: '{TESTING_CA_DIR}/ca-chain.cert.pem'
        """
    )
    setup = KeyFileConfigWrapper(cli_config).get_pkcs12_config('foo')
    with pytest.raises(ConfigurationError):
        KeyFileConfigWrapper(cli_config).get_pkcs12_config('bar')

    assert len(setup.other_certs) == 2

    signer = signer_from_p12_config(setup)
    _signer_sanity_check(signer)


def test_read_pkcs12_config_null_pw():
    cli_config = _parse_cli_config(
        f"""
        pkcs12-setups:
            foo:
                pfx-file: '{TESTING_CA_DIR}/interm/signer1.pfx'
                other-certs: '{TESTING_CA_DIR}/ca-chain.cert.pem'
                pfx-passphrase: null
        """
    )
    setup = KeyFileConfigWrapper(cli_config).get_pkcs12_config('foo')
    assert len(setup.other_certs) == 2

    signer = signer_from_p12_config(setup)
    _signer_sanity_check(signer)


def test_read_pemder_config():
    cli_config = _parse_cli_config(
        f"""
        pemder-setups:
            foo:
                key-file: '{CRYPTO_DATA_DIR}/keys-rsa/signer.key.pem'
                cert-file: '{TESTING_CA_DIR}/interm/signer1.cert.pem'
                other-certs: '{TESTING_CA_DIR}/ca-chain.cert.pem'
                key-passphrase: secret
        """
    )
    setup = KeyFileConfigWrapper(cli_config).get_pemder_config('foo')
    with pytest.raises(ConfigurationError):
        KeyFileConfigWrapper(cli_config).get_pemder_config('bar')

    assert len(setup.other_certs) == 2

    signer = signer_from_pemder_config(setup)
    _signer_sanity_check(signer)


def test_read_pemder_config_wrong_passphrase():
    cli_config = _parse_cli_config(
        f"""
        pemder-setups:
            foo:
                key-file: '{CRYPTO_DATA_DIR}/keys-rsa/signer.key.pem'
                cert-file: '{TESTING_CA_DIR}/interm/signer1.cert.pem'
                key-passphrase: "this passphrase is wrong"
        """
    )
    setup = KeyFileConfigWrapper(cli_config).get_pemder_config('foo')
    with pytest.raises(ConfigurationError):
        signer_from_pemder_config(setup)


def test_read_pemder_config_missing_passphrase():
    cli_config = _parse_cli_config(
        f"""
        pemder-setups:
            foo:
                key-file: '{CRYPTO_DATA_DIR}/keys-rsa/signer.key.pem'
                cert-file: '{TESTING_CA_DIR}/interm/signer1.cert.pem'
        """
    )
    setup = KeyFileConfigWrapper(cli_config).get_pemder_config('foo')
    with pytest.raises(ConfigurationError):
        signer_from_pemder_config(setup)

    signer_from_pemder_config(setup, provided_key_passphrase=b'secret')


def test_read_pkcs12_config_wrong_passphrase():
    cli_config = _parse_cli_config(
        f"""
        pkcs12-setups:
            foo:
                pfx-file: '{TESTING_CA_DIR}/interm/signer1.pfx'
                other-certs: '{TESTING_CA_DIR}/ca-chain.cert.pem'
                pfx-passphrase: "this passphrase is wrong"
        """
    )
    setup = KeyFileConfigWrapper(cli_config).get_pkcs12_config('foo')
    with pytest.raises(ConfigurationError):
        signer_from_p12_config(setup)


@pytest.mark.parametrize(
    'cfg_str,expected_result',
    [
        (
            "x-align: left",
            layout.SimpleBoxLayoutRule(
                x_align=layout.AxisAlignment.ALIGN_MIN,
                y_align=layout.AxisAlignment.ALIGN_MID,
            ),
        ),
        (
            "y-align: bottom",
            layout.SimpleBoxLayoutRule(
                x_align=layout.AxisAlignment.ALIGN_MID,
                y_align=layout.AxisAlignment.ALIGN_MIN,
            ),
        ),
        (
            f"""
        y-align: bottom
        x-align: mid
        margins:
            left: 10
            right: 10
        """,
            layout.SimpleBoxLayoutRule(
                x_align=layout.AxisAlignment.ALIGN_MID,
                y_align=layout.AxisAlignment.ALIGN_MIN,
                margins=layout.Margins(left=10, right=10),
            ),
        ),
        (
            f"""
        y-align: bottom
        x-align: mid
        margins: [10, 10, 0, 0]
        """,
            layout.SimpleBoxLayoutRule(
                x_align=layout.AxisAlignment.ALIGN_MID,
                y_align=layout.AxisAlignment.ALIGN_MIN,
                margins=layout.Margins(left=10, right=10),
            ),
        ),
        (
            f"""
        y-align: bottom
        x-align: mid
        inner-content-scaling: none
        """,
            layout.SimpleBoxLayoutRule(
                x_align=layout.AxisAlignment.ALIGN_MID,
                y_align=layout.AxisAlignment.ALIGN_MIN,
                inner_content_scaling=layout.InnerScaling.NO_SCALING,
            ),
        ),
        (
            "inner-content-scaling: stretch-to-fit",
            layout.SimpleBoxLayoutRule(
                x_align=layout.AxisAlignment.ALIGN_MID,
                y_align=layout.AxisAlignment.ALIGN_MID,
                inner_content_scaling=layout.InnerScaling.STRETCH_TO_FIT,
            ),
        ),
    ],
)
def test_read_simple_layout_config(cfg_str, expected_result):
    config_dict = yaml.safe_load(cfg_str)
    result = layout.SimpleBoxLayoutRule.from_config(config_dict)
    assert result == expected_result


@pytest.mark.parametrize(
    'cfg_str,error',
    [
        ("x-align: bottom", "is not a valid horizontal"),
        ("y-align: right", "is not a valid vertical"),
        ("inner-content-scaling: foobar", "is not a valid inner scaling"),
    ],
)
def test_read_simple_layout_config_failures(cfg_str, error):
    config_dict = yaml.safe_load(cfg_str)
    with pytest.raises(ConfigurationError, match=error):
        layout.SimpleBoxLayoutRule.from_config(config_dict)


@dataclass(frozen=True)
class DemoConfigurableA(ConfigurableMixin):
    field1: int
    field2: Iterable[int]
    field3: Optional[int] = None
    field4: Optional[Iterable[int]] = None
    field5: Union[str, int] = 'abc'
    field6: Union[str, int, None] = None


@dataclass(frozen=True)
class DemoConfigurableB(ConfigurableMixin):
    some_field: Optional[DemoConfigurableA] = None


@pytest.mark.parametrize(
    'cfg_str, expected_field_val',
    [
        (
            """
            some_field:
                field1: 1
                field2: [1,2]
            """,
            DemoConfigurableA(field1=1, field2=[1, 2]),
        ),
        (
            """
            some_field:
                field1: 1
                field2: [1,2]
                field3: 5
            """,
            DemoConfigurableA(field1=1, field2=[1, 2], field3=5),
        ),
        (
            """
            some_field:
                field1: 1
                field2: [1,2]
                field3: 5
                field4: [6,7,8]
            """,
            DemoConfigurableA(
                field1=1, field2=[1, 2], field3=5, field4=[6, 7, 8]
            ),
        ),
        (
            """
            some_field:
                field1: 1
                field2: [1,2]
                field5: xyz
            """,
            DemoConfigurableA(field1=1, field2=[1, 2], field5='xyz'),
        ),
        (
            """
            some_field:
                field1: 1
                field2: [1,2]
                field5: 8
                field6: xyz
            """,
            DemoConfigurableA(field1=1, field2=[1, 2], field5=8, field6='xyz'),
        ),
        ("{}", None),
    ],
)
def test_configurable_recurse(cfg_str, expected_field_val):
    config_dict = yaml.safe_load(cfg_str)
    b = DemoConfigurableB.from_config(config_dict)
    assert b.some_field == expected_field_val


@pytest.mark.parametrize(
    "cfg_str",
    [
        """
    field2: [1,2]
    field3: 5
    """,
        """
    field1: 1
    field3: 5
    """,
        """
    field3: 5
    """,
        "{}",
    ],
)
def test_enforce_required(cfg_str):
    config_dict = yaml.safe_load(cfg_str)
    with pytest.raises(ConfigurationError, match="Missing required key"):
        DemoConfigurableA.from_config(config_dict)


def test_enforce_required_recursive():
    config_dict = yaml.safe_load(
        """
        some_field:
            field2: [1,2]
        """
    )
    with pytest.raises(
        ConfigurationError, match="Error while processing configurable field"
    ):
        DemoConfigurableB.from_config(config_dict)


def test_default_stamp_style_fetch():
    # regression test for fetching the default stamp style if not explicitly
    # defined

    config_string = f"""
    validation-contexts:
        default:
            trust: '{TESTING_CA_DIR}/root/root.cert.pem'
            other-certs: '{TESTING_CA_DIR}/ca-chain.cert.pem'
    """

    cli_config: config.CLIConfig = _parse_cli_config(config_string)

    result = cli_config.get_stamp_style(None)
    from pyhanko.sign import DEFAULT_SIGNING_STAMP_STYLE

    assert result == DEFAULT_SIGNING_STAMP_STYLE



"""
Module defining pyHanko's standard difference policy implementation.
"""

import logging
from collections import defaultdict
from typing import Iterator, List, Optional, Union

from pyhanko.pdf_utils import generic, misc
from pyhanko.pdf_utils.reader import HistoricalResolver, PdfFileReader
from pyhanko.sign.fields import FieldMDPSpec, MDPPerm

from .form_rules_api import FormUpdatingRule
from .policy_api import (
    DiffPolicy,
    DiffResult,
    ModificationLevel,
    SuspiciousModification,
)
from .rules.file_structure_rules import (
    CatalogModificationRule,
    ObjectStreamRule,
    XrefStreamRule,
)
from .rules.form_field_rules import (
    DSSCompareRule,
    GenericFieldModificationRule,
    SigFieldCreationRule,
    SigFieldModificationRule,
)
from .rules.metadata_rules import DocInfoRule, MetadataUpdateRule
from .rules_api import (
    ApprovalType,
    Context,
    QualifiedWhitelistRule,
    ReferenceUpdate,
)

logger = logging.getLogger(__name__)


__all__ = [
    'StandardDiffPolicy',
    'DEFAULT_DIFF_POLICY',
    'NO_CHANGES_DIFF_POLICY',
]


def _find_orphans(hist_rev: HistoricalResolver):
    """
    Within a revision, find new refs that can't be reached from refs in the
    older ones.
    """

    # Note: this function assumes that there is no shady behaviour with older
    #  revisions referring to as-of-yet-undefined references in future
    #  revisions.
    # TODO I might want to put a failsafe in the PdfFileReader class's
    #  dereferencing logic to prevent that.

    # This assumption makes finding orphans relatively cheap: we only need to
    # pull up the dependencies of the older objects that were overwritten
    # in this exact revision, and we only have to recurse into branches that
    # pass through new objects themselves.

    new_refs = hist_rev.explicit_refs_in_revision()

    previous = hist_rev.reader.get_historical_resolver(hist_rev.revision - 1)

    # These are newly updated refs that already existed in older revisions.
    #  We want to know which of the new refs are reachable from one of these.
    updated_old_refs = set()
    # The candidate orphans are all the others
    candidate_orphans = set()
    for ref in new_refs:
        if previous.is_ref_available(ref):
            # ref didn't exist in previous revision
            candidate_orphans.add(ref)
        else:
            updated_old_refs.add(ref)

    def _objs_to_check() -> Iterator[generic.PdfObject]:
        # check the trailer too!
        yield hist_rev.trailer_view
        for _ref in updated_old_refs:
            # take care to return the historical value here
            yield hist_rev(_ref)

    obj_iter = _objs_to_check()
    while candidate_orphans:
        try:
            obj = next(obj_iter)
        except StopIteration:
            break
        candidate_orphans -= hist_rev.collect_dependencies(
            obj, since_revision=hist_rev.revision
        )
    return candidate_orphans


def _is_id(old_object: generic.PdfObject, new_object: generic.PdfObject):
    primitives = (
        generic.NumberObject,
        generic.FloatObject,
        generic.BooleanObject,
        generic.NameObject,
        generic.NullObject,
        generic.IndirectObject,
    )
    for prim in primitives:
        if isinstance(old_object, prim):
            return isinstance(new_object, prim) and new_object == old_object

    strings = (generic.TextStringObject, generic.ByteStringObject)
    if isinstance(old_object, strings):
        return (
            isinstance(new_object, strings)
            and old_object.original_bytes == new_object.original_bytes
        )

    if isinstance(old_object, generic.ArrayObject):
        return (
            isinstance(new_object, generic.ArrayObject)
            and len(new_object) == len(old_object)
            and all(_is_id(x, y) for x, y in zip(old_object, new_object))
        )

    if isinstance(old_object, generic.StreamObject):
        # fallthrough to dict case if this check passes
        if not (
            isinstance(new_object, generic.StreamObject)
            and new_object.encoded_data == old_object.encoded_data
        ):
            return False
    if isinstance(old_object, generic.DictionaryObject):
        if not (
            isinstance(new_object, generic.DictionaryObject)
            and new_object.keys() == old_object.keys()
        ):
            return False

        return all(
            _is_id(old_object.raw_get(k), new_object.raw_get(k))
            for k in new_object.keys()
        )
    raise NotImplementedError


class StandardDiffPolicy(DiffPolicy):
    """
    Run a list of rules to analyse the differences between two revisions.

    :param global_rules:
        The :class:`.QualifiedWhitelistRule` objects encoding the rules to
        apply.
    :param form_rule:
        The :class:`.FormUpdatingRule` that adjudicates changes to form fields
        and their values.
    :param reject_object_freeing:
        Always fail revisions that free objects that existed prior to signing.

        .. note::
            PyHanko resolves freed references to the ``null`` object in PDF,
            and a freeing instruction in a cross-reference section is
            always registered as a change that needs to be approved, regardless
            of the value of this setting.

            It is theoretically possible for a rule to permit deleting content,
            in which case allowing objects to be freed might be reasonable.
            That said, pyHanko takes the conservative default position to reject
            all object freeing instructions as suspect.
    :param ignore_orphaned_objects:
        Some PDF writers create objects that aren't used anywhere (tsk tsk).
        Since those don't affect the "actual" document content, they can usually
        be ignored. If ``True``, newly created orphaned objects will be
        cleared at level :attr:`.ModificationLevel.LTA_UPDATES`.
        Default is ``True``.
    :param ignore_orphaned_objects:
        Some PDF writers overwrite objects with identical copies.
        Pointless and annoying, but also more or less harmless.
    """

    def __init__(
        self,
        global_rules: List[QualifiedWhitelistRule],
        form_rule: Optional[FormUpdatingRule],
        reject_object_freeing=True,
        ignore_orphaned_objects=True,
        ignore_identical_objects=True,
    ):
        self.global_rules = global_rules
        self.form_rule = form_rule
        self.reject_object_freeing = reject_object_freeing
        self.ignore_orphaned_objects = ignore_orphaned_objects
        self.ignore_identical_objects = ignore_identical_objects

    def apply(
        self,
        old: HistoricalResolver,
        new: HistoricalResolver,
        field_mdp_spec: Optional[FieldMDPSpec] = None,
        doc_mdp: Optional[MDPPerm] = None,
    ) -> DiffResult:
        if doc_mdp == MDPPerm.ANNOTATE:
            logger.warning(
                "StandardDiffPolicy was not designed to support "
                "DocMDP level 3 (MDPPerm.ANNOTATE). Unexpected validation "
                "results may occur."
            )

        if self.reject_object_freeing:
            freed = new.refs_freed_in_revision()
            if freed:
                raise SuspiciousModification(
                    f"The refs {freed} were freed in the revision provided. "
                    "The configured difference analysis policy does not allow "
                    "object freeing."
                )
        # we need to verify that there are no xrefs in the revision's xref table
        # other than the ones we can justify.
        new_xrefs = new.explicit_refs_in_revision()

        explained = defaultdict(set)

        # prepare LUT for refs that are used multiple times in the old revision
        # (this is a very expensive operation, since it reads all objects in
        #  the signed revision)
        def _init_multi_lut():
            old._load_reverse_xref_cache()
            for new_ref in new_xrefs:
                usages = old._get_usages_of_ref(new_ref)
                if usages:
                    contexts = {Context.from_absolute(old, p) for p in usages}
                    yield new_ref, (ModificationLevel.NONE, contexts)

        # orphaned objects are cleared at LTA update level
        if self.ignore_orphaned_objects:
            for _ref in _find_orphans(new):
                explained[ModificationLevel.LTA_UPDATES].add(_ref)

        # This table records all the overridden refs that already existed
        # in the old revision, together with the different ways they can be
        # reached from the document trailer.
        # Unlike fresh refs, these need to be cleared together with the paths
        # through which they are accessed.
        old_usages_to_clear = dict(_init_multi_lut())

        def ingest_ref(_level: ModificationLevel, _upd: ReferenceUpdate):
            ref = _upd.updated_ref
            try:
                current_max_level, usages = old_usages_to_clear[ref]
                upd_type = _upd.approval_type
                if upd_type == ApprovalType.BLANKET_APPROVE:
                    # approve all usages at once
                    usages = set()
                elif upd_type == ApprovalType.APPROVE_RELATIVE_CONTEXT:
                    # approve all usages that match the relative context,
                    # keep the rest
                    usages = set(
                        ctx
                        for ctx in usages
                        if ctx.relative_view != _upd.context_checked
                    )
                else:
                    # Last case: path-level approval
                    # remove the path that has just been cleared from
                    # the checklist
                    usages.discard(_upd.context_checked)
                # bump the modification level for this reference if necessary
                _level = max(current_max_level, _level)
                old_usages_to_clear[ref] = _level, usages
                if usages:
                    # not all paths/usages have been cleared, so we can't
                    # approve the reference yet
                    return
            except KeyError:
                pass
            explained[_level].add(ref)

        for rule in self.global_rules:
            for level, upd in rule.apply_qualified(old, new):
                ingest_ref(level, upd)

        changed_form_fields = set()

        if self.form_rule:
            form_changes = self.form_rule.apply(old, new)

            def is_locked(fq_name):
                return field_mdp_spec is not None and field_mdp_spec.is_locked(
                    fq_name
                )

            for level, fu in form_changes:
                ingest_ref(level, fu)
                field_name = fu.field_name
                if field_name is not None and not fu.valid_when_locked:
                    if is_locked(field_name):
                        raise SuspiciousModification(
                            f"Update of {fu.updated_ref} is not allowed "
                            f"because the form field {field_name} is locked."
                        )
                    changed_form_fields.add(field_name)
                if doc_mdp is not None and not fu.valid_when_certifying:
                    raise SuspiciousModification(
                        f"Update of {fu.updated_ref} is only allowed "
                        f"after an approval signature, not a certification "
                        f"signature."
                    )

        unexplained_lta = new_xrefs - explained[ModificationLevel.LTA_UPDATES]
        unexplained_formfill = (
            unexplained_lta - explained[ModificationLevel.FORM_FILLING]
        )
        unexplained_annot = (
            unexplained_formfill - explained[ModificationLevel.ANNOTATIONS]
        )

        xref_cache = old.reader.xrefs
        if self.ignore_identical_objects:
            identical_objs = set(
                unex_ref
                for unex_ref in unexplained_lta
                if xref_cache.get_historical_ref(unex_ref, old.revision)
                is not None
                and _is_id(old(unex_ref), new(unex_ref))
            )
            if identical_objs:
                logger.debug(
                    f"Found identical overridden objects between revisions "
                    f"{old.revision} and {new.revision}; following no-op "
                    f"changes will be ignored: {identical_objs}"
                )
                unexplained_annot.difference_update(identical_objs)
                unexplained_formfill.difference_update(identical_objs)
                unexplained_lta.difference_update(identical_objs)

        if unexplained_annot:
            msg = misc.LazyJoin(
                '\n',
                (
                    '%s:%s...' % (repr(x), repr(x.get_object())[:300])
                    for x in unexplained_annot
                ),
            )
            logger.debug(
                "Unexplained xrefs in revision %d:\n%s", new.revision, msg
            )
            unexplained_overrides = [
                f" - {repr(ref)} is also used in "
                f"{', '.join(str(p) for p in paths_remaining)} in the prior "
                f"revision."
                for ref, (_, paths_remaining) in old_usages_to_clear.items()
                if paths_remaining
            ]
            err_msg = (
                f"There are unexplained xrefs in revision {new.revision}: "
                f"{', '.join(repr(x) for x in unexplained_annot)}."
            )
            if unexplained_overrides:
                unchecked_paths_msg = (
                    f"Some objects from revision {old.revision} were replaced "
                    f"in revision {new.revision} without precise "
                    "justification:\n" + '\n'.join(unexplained_overrides)
                )
                err_msg = "%s\n%s" % (err_msg, unchecked_paths_msg)
                logger.debug(unchecked_paths_msg)

            raise SuspiciousModification(err_msg)
        elif unexplained_formfill:
            level = ModificationLevel.ANNOTATIONS
        elif unexplained_lta:
            level = ModificationLevel.FORM_FILLING
        else:
            level = ModificationLevel.LTA_UPDATES

        return DiffResult(
            modification_level=level, changed_form_fields=changed_form_fields
        )

    def review_file(
        self,
        reader: PdfFileReader,
        base_revision: Union[int, HistoricalResolver],
        field_mdp_spec: Optional[FieldMDPSpec] = None,
        doc_mdp: Optional[MDPPerm] = None,
    ) -> Union[DiffResult, SuspiciousModification]:
        """
        Implementation of :meth:`.DiffPolicy.review_file` that reviews
        each intermediate revision between the base revision and the current one
        individually.
        """

        changed_form_fields = set()

        rev_count = reader.xrefs.total_revisions
        current_max = ModificationLevel.NONE
        if isinstance(base_revision, int):
            base_rev_resolver = reader.get_historical_resolver(base_revision)
            base_revision_no = base_revision
        else:
            base_rev_resolver = base_revision
            base_revision_no = base_rev_resolver.revision

        # Note: there's a pragmatic reason why we iterate over all revisions
        # instead of just asking for all updated objects between the signed
        # revision and the most recent one:
        #
        # The effect of intermediate updates may not be detectable anymore in
        # the most recent version, so if we'd consolidate all checks into one,
        # we would have no way to tell whether or not the objects created
        # (and later forgotten) by these intermediate revisions actually
        # constituted legitimate changes.
        # (see the test_pades_revinfo tests for examples where this applies)
        #
        # Until we have a reference counter (which comes with its own
        # performance problems that may or may not be worse), I don't really
        # see a good way around this issue other than diffing every intermediate
        # version separately.
        for revision in range(base_revision_no + 1, rev_count):
            try:
                diff_result = self.apply(
                    old=base_rev_resolver,
                    new=reader.get_historical_resolver(revision),
                    field_mdp_spec=field_mdp_spec,
                    doc_mdp=doc_mdp,
                )
            except SuspiciousModification as e:
                logger.warning(
                    'Error in diff operation between revision '
                    f'{base_revision_no} and {revision}',
                    exc_info=e,
                )
                return e
            current_max = max(current_max, diff_result.modification_level)
            changed_form_fields |= diff_result.changed_form_fields
        return DiffResult(current_max, changed_form_fields)


DEFAULT_DIFF_POLICY = StandardDiffPolicy(
    global_rules=[
        CatalogModificationRule(),
        DocInfoRule().as_qualified(ModificationLevel.LTA_UPDATES),
        XrefStreamRule().as_qualified(ModificationLevel.LTA_UPDATES),
        ObjectStreamRule().as_qualified(ModificationLevel.LTA_UPDATES),
        DSSCompareRule().as_qualified(ModificationLevel.LTA_UPDATES),
        MetadataUpdateRule().as_qualified(ModificationLevel.LTA_UPDATES),
    ],
    form_rule=FormUpdatingRule(
        field_rules=[
            SigFieldCreationRule(),
            SigFieldModificationRule(),
            GenericFieldModificationRule(),
        ],
    ),
)
"""
Default :class:`.DiffPolicy` implementation.

This policy includes the following rules, all with the default settings.
The unqualified rules in the list all have their updates qualified at
level :class:`~.ModificationLevel.LTA_UPDATES`.

* :class:`.CatalogModificationRule`,
* :class:`.DocInfoRule`,
* :class:`.ObjectStreamRule`,
* :class:`.XrefStreamRule`,
* :class:`.DSSCompareRule`,
* :class:`.MetadataUpdateRule`.
* :class:`.FormUpdatingRule`, with the following field rules:

    * :class:`.SigFieldCreationRule`,
    * :class:`.SigFieldModificationRule`,
    * :class:`.GenericFieldModificationRule`.
"""


NO_CHANGES_DIFF_POLICY = StandardDiffPolicy(global_rules=[], form_rule=None)
"""
:class:`.DiffPolicy` implementation that does not provide any rules,
and will therefore simply reject all changes.
"""



import itertools
import os
from datetime import datetime, timezone
from io import BytesIO

import pytest
import tzlocal
from freezegun import freeze_time
from pyhanko.pdf_utils import generic
from pyhanko.pdf_utils.generic import EncryptedObjAccess
from pyhanko.pdf_utils.incremental_writer import IncrementalPdfFileWriter
from pyhanko.pdf_utils.metadata import model, xmp_xml
from pyhanko.pdf_utils.misc import StringWithLanguage
from pyhanko.pdf_utils.reader import HistoricalResolver, PdfFileReader
from pyhanko.pdf_utils.writer import PdfFileWriter, copy_into_new_writer

from .samples import MINIMAL, PDF_DATA_DIR, VECTOR_IMAGE_PDF

try:
    import zoneinfo
except ImportError:
    # noinspection PyUnresolvedReferences
    from backports import zoneinfo


def test_no_meta_view_fails_gracefully():
    r = PdfFileReader(BytesIO(MINIMAL))
    assert r.document_meta_view == model.DocumentMetadata()


@freeze_time(datetime(2022, 9, 7, tzinfo=tzlocal.get_localzone()))
def test_copy_into_new_writer_sets_info():
    r = PdfFileReader(BytesIO(MINIMAL))
    w = copy_into_new_writer(r)
    out = BytesIO()
    w.write(out)

    r = PdfFileReader(out)
    meta = r.document_meta_view
    assert meta.last_modified == datetime.now(tz=tzlocal.get_localzone())

    assert 'pyHanko' in r.trailer_view['/Info']['/Producer']


def test_date_parse_failure():
    r = PdfFileReader(BytesIO(MINIMAL))
    writer_kwargs = {
        'info': generic.DictionaryObject(
            {
                generic.pdf_name('/Title'): generic.pdf_string(
                    'a failure test'
                ),
                generic.pdf_name('/CreationDate'): generic.pdf_string(
                    'this makes no sense'
                ),
            }
        )
    }
    w = copy_into_new_writer(r, writer_kwargs=writer_kwargs)
    w._update_meta = lambda: None
    out = BytesIO()
    w.write(out)

    r = PdfFileReader(out)
    meta = r.document_meta_view
    assert meta.created is None
    assert meta.title == "a failure test"


@freeze_time(datetime(2022, 9, 7, tzinfo=tzlocal.get_localzone()))
def test_incremental_update_meta_view():
    w = IncrementalPdfFileWriter(BytesIO(VECTOR_IMAGE_PDF))
    assert w.document_meta_view.created.year == 2020
    w.document_meta.created = datetime.now()
    assert w.document_meta_view.created.year == 2022


@freeze_time(datetime(2022, 9, 7, tzinfo=tzlocal.get_localzone()))
@pytest.mark.parametrize('writer_type', ('fresh', 'from_data', 'incremental'))
def test_writer_meta_view_does_not_persist_changes(writer_type):
    exp_value = datetime(
        2020, 9, 5, 19, 30, 57, tzinfo=zoneinfo.ZoneInfo('CET')
    )
    if writer_type == 'fresh':
        w = PdfFileWriter()
        exp_value = None
    elif writer_type == 'from_data':
        w = copy_into_new_writer(PdfFileReader(BytesIO(VECTOR_IMAGE_PDF)))
    else:
        w = IncrementalPdfFileWriter(BytesIO(VECTOR_IMAGE_PDF))
    assert w.document_meta_view.created == exp_value
    w.document_meta_view.created = datetime.now()
    assert w.document_meta_view.created == exp_value


@freeze_time(datetime(2022, 9, 7, tzinfo=tzlocal.get_localzone()))
@pytest.mark.parametrize(
    'writer_type,meta_dict,ver',
    list(
        itertools.product(
            ('fresh', 'from_data', 'incremental', 'in_place'),
            [
                {'title': 'Test test'},
                {'author': 'John Doe'},
                {'keywords': ['foo', 'bar', 'baz']},
                {'created': datetime(2022, 9, 7, tzinfo=timezone.utc)},
                {'author': 'John Doe', 'subject': 'Blah blah blah'},
            ],
            ("pdf1.7", "pdf2.0"),
        )
    ),
)
def test_metadata_info_round_trip(writer_type, meta_dict: dict, ver):
    out = BytesIO()
    if writer_type == 'fresh':
        w = PdfFileWriter()
    elif writer_type == 'from_data':
        w = copy_into_new_writer(PdfFileReader(BytesIO(MINIMAL)))
    else:
        w = IncrementalPdfFileWriter(BytesIO(MINIMAL))
    w.output_version = (1, 7) if ver == "pdf1.7" else (2, 0)

    w.update_root()

    for k, v in meta_dict.items():
        setattr(w.document_meta, k, v)

    if writer_type == 'in_place':
        w.write_in_place()
        out = w.prev.stream
    else:
        w.write(out)

    r = PdfFileReader(out)
    meta_dict['last_modified'] = datetime.now(tz=tzlocal.get_localzone())
    if ver == "pdf2.0":
        assert '/Info' not in r.trailer_view
    else:
        assert r.document_meta_view == model.DocumentMetadata(**meta_dict)


@pytest.mark.parametrize(
    'writer_type,ver',
    list(
        itertools.product(
            ('fresh', 'from_data', 'incremental', 'in_place'), ((1, 7), (2, 0))
        )
    ),
)
def test_metadata_info_with_language_round_trip(writer_type, ver):
    title = StringWithLanguage('Test test', lang_code='en', country_code='US')
    out = BytesIO()
    if writer_type == 'fresh':
        w = PdfFileWriter()
    elif writer_type == 'from_data':
        w = copy_into_new_writer(PdfFileReader(BytesIO(MINIMAL)))
    else:
        w = IncrementalPdfFileWriter(BytesIO(MINIMAL))
    w.output_version = ver

    w.update_root()
    w.document_meta.title = title

    if writer_type == 'in_place':
        w.write_in_place()
        out = w.prev.stream
    else:
        w.write(out)

    r = PdfFileReader(out)
    if '/Metadata' in r.root:
        # language info preserved
        assert r.document_meta_view.title == title
    else:
        # language info not preserved
        assert r.document_meta_view.title == title.value


# Example from ISO 16684-1:20121 7.9.2.3

# noinspection HttpUrlsUsage
XMP_WITH_PARSETYPE_RESOURCE = """
<rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#"
         xmlns:xmpTPg="http://ns.adobe.com/xap/1.0/t/pg/"
         xmlns:stDim="http://ns.adobe.com/xap/1.0/sType/Dimensions#"
         xmlns:xmp="http://ns.adobe.com/xap/1.0/"
         xmlns:xe="http://ns.adobe.com/xmp-example/">
    <rdf:Description rdf:about="">
        <xmpTPg:MaxPageSize rdf:parseType="Resource">
            <stDim:h>11.0</stDim:h>
            <stDim:w>8.5</stDim:w>
            <stDim:unit>inch</stDim:unit>
        </xmpTPg:MaxPageSize>
        <xmp:BaseURL rdf:parseType="Resource">
            <rdf:value rdf:resource="https://www.adobe.com/"/>
            <xe:qualifier>artificial example</xe:qualifier>
        </xmp:BaseURL>
   </rdf:Description>
</rdf:RDF>
"""


# noinspection HttpUrlsUsage
def test_xmp_parsetype_resource():
    inp = BytesIO(XMP_WITH_PARSETYPE_RESOURCE.encode('utf8'))
    result = xmp_xml.parse_xmp(inp)[0]

    n_max_page_size = model.ExpandedName(
        ns="http://ns.adobe.com/xap/1.0/t/pg/", local_name="MaxPageSize"
    )
    n_unit = model.ExpandedName(
        ns="http://ns.adobe.com/xap/1.0/sType/Dimensions#", local_name="unit"
    )
    n_base_url = model.ExpandedName(ns=model.NS['xmp'], local_name="BaseURL")
    n_xe_qualifier = model.ExpandedName(
        ns="http://ns.adobe.com/xmp-example/", local_name="qualifier"
    )
    max_page_size = result[n_max_page_size].value
    assert max_page_size[n_unit].value == 'inch'

    base_url_val = result[n_base_url]
    assert base_url_val.qualifiers[n_xe_qualifier].value == 'artificial example'
    assert base_url_val.value == model.XmpUri("https://www.adobe.com/")


def test_ensure_uri_is_rdf_resource():
    n_base_url = model.ExpandedName(ns=model.NS['xmp'], local_name="BaseURL")
    root = model.XmpStructure.of(
        (n_base_url, model.XmpValue(model.XmpUri("https://example.com")))
    )
    out = BytesIO()
    xmp_xml.serialise_xmp([root], out)
    assert b'rdf:resource="https://example.com"' in out.getvalue()


# noinspection HttpUrlsUsage
@freeze_time('2022-09-10')
def test_incremental_update_doc_with_xmp():
    with open(
        os.path.join(PDF_DATA_DIR, "minimal-pdf-ua-and-a.pdf"), 'rb'
    ) as inf:
        w = IncrementalPdfFileWriter(inf)
        w.root['/Foo'] = generic.NameObject('/Bar')
        w.document_meta.subject = "Update test"
        out = BytesIO()
        w.write(out)
    r = PdfFileReader(out)
    xmp: model.XmpStructure = r.root['/Metadata'].xmp[0]
    assert xmp[model.DC_DESCRIPTION].value.entries[0].value == "Update test"
    pdfa_conformance = model.ExpandedName(
        "http://www.aiim.org/pdfa/ns/id/", "conformance"
    )
    assert xmp[pdfa_conformance].value == "B"

    assert HistoricalResolver(r, 0).document_meta_view.subject is None


# noinspection HttpUrlsUsage
@freeze_time('2022-09-10')
def test_rewrite_update_doc_with_xmp():
    with open(
        os.path.join(PDF_DATA_DIR, "minimal-pdf-ua-and-a.pdf"), 'rb'
    ) as inf:
        w = copy_into_new_writer(PdfFileReader(inf))
        w.document_meta.subject = "Update test"
        out = BytesIO()
        w.write(out)
    r = PdfFileReader(out)
    xmp: model.XmpStructure = r.root['/Metadata'].xmp[0]
    assert xmp[model.DC_DESCRIPTION].value.entries[0].value == "Update test"
    pdfa_conformance = model.ExpandedName(
        "http://www.aiim.org/pdfa/ns/id/", "conformance"
    )
    assert xmp[pdfa_conformance].value == "B"


@freeze_time('2022-09-10')
def test_meta_view_from_xmp():
    with open(
        os.path.join(PDF_DATA_DIR, "minimal-pdf-ua-and-a.pdf"), 'rb'
    ) as inf:
        r = PdfFileReader(inf)
        # wipe out the info dict just because
        r.trailer['/Info'] = generic.DictionaryObject()
        assert r.document_meta_view.title == StringWithLanguage(
            value="Test document", lang_code="DEFAULT"
        )


def test_upgrade_pdf2_no_info_dict():
    buf = BytesIO(MINIMAL)
    w = IncrementalPdfFileWriter(buf)
    w.output_version = (2, 0)
    w.root['/Foo'] = generic.NameObject('/Bar')
    w.document_meta.title = "Test document"
    w.document_meta.author = "John Doe"
    w.document_meta.subject = "Test subject"
    w.document_meta.keywords = ["these", "are", "key", "words"]
    w.write_in_place()

    r = PdfFileReader(buf)
    assert '/Info' not in r.trailer_view
    assert r.document_meta_view.title.value == "Test document"
    assert r.document_meta_view.author == "John Doe"
    assert r.document_meta_view.subject.value == "Test subject"
    assert r.document_meta_view.keywords == ["these", "are", "key", "words"]

    # historical resolver only looks at info dict
    assert HistoricalResolver(r, 0).document_meta_view.subject is None


# noinspection HttpUrlsUsage
def test_add_extra_xmp():
    buf = BytesIO(MINIMAL)
    w = IncrementalPdfFileWriter(buf)
    w.root['/Foo'] = generic.NameObject('/Bar')
    w.document_meta.title = "Test document"

    n_base_url = model.ExpandedName(ns=model.NS['xmp'], local_name="BaseURL")
    n_xe_qualifier = model.ExpandedName(
        ns="http://ns.adobe.com/xmp-example/", local_name="qualifier"
    )
    extra = model.XmpStructure.of(
        (
            n_base_url,
            model.XmpValue(
                model.XmpUri("https://example.com/"),
                qualifiers=model.Qualifiers.of(
                    (n_xe_qualifier, model.XmpValue('artificial example'))
                ),
            ),
        ),
    )
    w.document_meta.xmp_extra = [extra]
    w.write_in_place()

    r = PdfFileReader(buf)
    assert '/Info' not in r.trailer_view
    assert r.document_meta_view.title.value == "Test document"

    base_url_val = r.root['/Metadata'].xmp[1][n_base_url]
    assert base_url_val.qualifiers[n_xe_qualifier].value == 'artificial example'
    assert base_url_val.value == model.XmpUri("https://example.com/")


# noinspection HttpUrlsUsage
def test_unmanaged_xmp():
    r = PdfFileReader(BytesIO(MINIMAL))
    w = copy_into_new_writer(r)
    n_base_url = model.ExpandedName(ns=model.NS['xmp'], local_name="BaseURL")
    n_xe_qualifier = model.ExpandedName(
        ns="http://ns.adobe.com/xmp-example/", local_name="qualifier"
    )
    extra = model.XmpStructure.of(
        (
            n_base_url,
            model.XmpValue(
                model.XmpUri("https://example.com/"),
                qualifiers=model.Qualifiers.of(
                    (n_xe_qualifier, model.XmpValue('artificial example'))
                ),
            ),
        ),
    )
    w.document_meta.title = "This should not be written"
    w.document_meta.xmp_unmanaged = True
    w.document_meta.xmp_extra = [extra]
    out = BytesIO()
    w.write(out)

    r = PdfFileReader(out)
    assert '/Info' not in r.trailer_view
    assert r.document_meta_view.title is None

    base_url_val = r.root['/Metadata'].xmp[1][n_base_url]
    assert base_url_val.qualifiers[n_xe_qualifier].value == 'artificial example'
    assert base_url_val.value == model.XmpUri("https://example.com/")


# noinspection HttpUrlsUsage
def test_unmanaged_xmp_does_not_affect_info():
    # use a PDF that already has an info dictionary
    r = PdfFileReader(BytesIO(VECTOR_IMAGE_PDF))
    w = copy_into_new_writer(r)
    n_base_url = model.ExpandedName(ns=model.NS['xmp'], local_name="BaseURL")
    n_xe_qualifier = model.ExpandedName(
        ns="http://ns.adobe.com/xmp-example/", local_name="qualifier"
    )
    extra = model.XmpStructure.of(
        (
            n_base_url,
            model.XmpValue(
                model.XmpUri("https://example.com/"),
                qualifiers=model.Qualifiers.of(
                    (n_xe_qualifier, model.XmpValue('artificial example'))
                ),
            ),
        ),
    )
    w.document_meta.title = "This should not be written"
    w.document_meta.xmp_unmanaged = True
    w.document_meta.xmp_extra = [extra]
    out = BytesIO()
    w.write(out)

    r = PdfFileReader(out)
    meta = r.document_meta_view
    assert meta.title is None
    # producer line should update
    assert 'pyHanko' in r.trailer_view['/Info']['/Producer']


# noinspection HttpUrlsUsage
XMP_WITH_RESOURCE_WITH_MULTIPLE_QUALS = """
<rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#"
         xmlns:dc="http://purl.org/dc/elements/1.1/"
         xmlns:xe="http://ns.adobe.com/xmp-example/">
    <rdf:Description rdf:about="">
        <dc:title rdf:parseType="Resource" xml:lang="en">
            <!-- NOTE: this is actually not valid in the DC schema -->
            <rdf:value>This is a test</rdf:value>
            <xe:qualifier>artificial example</xe:qualifier>
        </dc:title>
   </rdf:Description>
</rdf:RDF>
"""


# noinspection HttpUrlsUsage
def test_xmp_with_resource_with_multiple_quals():
    inp = BytesIO(XMP_WITH_RESOURCE_WITH_MULTIPLE_QUALS.encode('utf8'))
    result = xmp_xml.parse_xmp(inp)[0]
    qual_ct = result[model.DC_TITLE].qualifiers.iter_quals(with_lang=True)
    assert len(list(qual_ct)) == 2

    n_xe_qualifier = model.ExpandedName(
        ns="http://ns.adobe.com/xmp-example/", local_name="qualifier"
    )

    assert result == model.XmpStructure.of(
        (
            model.DC_TITLE,
            model.XmpValue(
                "This is a test",
                model.Qualifiers.of(
                    (model.XML_LANG, model.XmpValue("en")),
                    (
                        n_xe_qualifier,
                        model.XmpValue(
                            "artificial example",
                            model.Qualifiers.lang_as_qual("en"),
                        ),
                    ),
                ),
            ),
        ),
        (model.RDF_ABOUT, model.XmpValue("")),
    )
    title_val = result[model.DC_TITLE]
    assert title_val.qualifiers[n_xe_qualifier].value == 'artificial example'
    assert xmp_xml.meta_from_xmp([result]).title == model.StringWithLanguage(
        "This is a test", lang_code="en"
    )


def test_xmp_str():
    inp = BytesIO(XMP_WITH_RESOURCE_WITH_MULTIPLE_QUALS.encode('utf8'))
    result = xmp_xml.parse_xmp(inp)[0]
    r = repr(result)
    assert "artificial" in r
    assert "lang': \'en\'" in r


# noinspection HttpUrlsUsage
XMP_WITH_BAG = """
<rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#"
         xmlns:xe="http://ns.adobe.com/xmp-example/">
    <rdf:Description rdf:about="">
        <xe:keywords rdf:parseType="Resource">
            <rdf:value>
                <rdf:Bag>
                    <rdf:li>apple</rdf:li>
                    <rdf:li>pear</rdf:li>
                    <rdf:li>banana</rdf:li>
                </rdf:Bag>
            </rdf:value>
            <xml:lang>en</xml:lang>
            <xe:qualifier>artificial example</xe:qualifier>
        </xe:keywords>
   </rdf:Description>
</rdf:RDF>
"""


# noinspection HttpUrlsUsage
def test_xmp_with_bag():
    inp = BytesIO(XMP_WITH_BAG.encode('utf8'))
    result = xmp_xml.parse_xmp(inp)[0]

    n_xe_keywords = model.ExpandedName(
        ns="http://ns.adobe.com/xmp-example/", local_name="keywords"
    )
    n_xe_qualifier = model.ExpandedName(
        ns="http://ns.adobe.com/xmp-example/", local_name="qualifier"
    )

    keywords_val = result[n_xe_keywords]
    assert keywords_val.qualifiers[n_xe_qualifier].value == 'artificial example'

    assert keywords_val.value == model.XmpArray.unordered(
        map(model.XmpValue, ['apple', 'banana', 'pear'])
    )


def test_xmp_ord_arr_comparison():
    o1 = model.XmpArray.ordered(
        map(model.XmpValue, ['apple', 'pear', 'banana'])
    )
    o2 = model.XmpArray.ordered(
        map(model.XmpValue, ['apple', 'banana', 'pear'])
    )
    a = model.XmpArray.alternative(
        map(model.XmpValue, ['apple', 'pear', 'banana'])
    )
    assert o1 != o2
    assert o1 != a


def test_xmp_unord_arr_comparison():
    u1 = model.XmpArray.unordered(
        map(model.XmpValue, ['apple', 'pear', 'banana'])
    )
    u2 = model.XmpArray.unordered(
        map(model.XmpValue, ['apple', 'banana', 'pear'])
    )
    assert u1 == u2


# noinspection HttpUrlsUsage
XMP_WITH_PARSETYPE_LIT_INVALID = """
<rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#"
         xmlns:xe="http://ns.adobe.com/xmp-example/">
    <rdf:Description rdf:about="">
        <xe:title rdf:parseType="Literal">bleh</xe:title>
   </rdf:Description>
</rdf:RDF>
"""


def test_xmp_parsetype_lit_not_supported():
    with pytest.raises(xmp_xml.XmpXmlProcessingError, match="Literal"):
        xmp_xml.parse_xmp(
            BytesIO(XMP_WITH_PARSETYPE_LIT_INVALID.encode('utf8'))
        )


# noinspection HttpUrlsUsage
XMP_WITH_INVALID_VALUE_FORM = """
<rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#"
         xmlns:xe="http://ns.adobe.com/xmp-example/">
    <rdf:Description rdf:about="">
        <xe:title><xe:zzz/></xe:title>
   </rdf:Description>
</rdf:RDF>
"""


def test_xmp_with_invalid_form():
    with pytest.raises(xmp_xml.XmpXmlProcessingError, match="value form"):
        xmp_xml.parse_xmp(BytesIO(XMP_WITH_INVALID_VALUE_FORM.encode('utf8')))


# noinspection HttpUrlsUsage
XMP_VALUE_FORM_MULTIPLE_CHILDREN = """
<rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#"
         xmlns:xe="http://ns.adobe.com/xmp-example/">
    <rdf:Description rdf:about="">
        <xe:title><xe:foo/><xe:bar/></xe:title>
   </rdf:Description>
</rdf:RDF>
"""


def test_xmp_value_form_multiple_children():
    with pytest.raises(
        xmp_xml.XmpXmlProcessingError, match="more than one child"
    ):
        xmp_xml.parse_xmp(
            BytesIO(XMP_VALUE_FORM_MULTIPLE_CHILDREN.encode('utf8'))
        )


XMP_NO_RDF_IN_XMPMETA = """
<x:xmpmeta xmlns:x="adobe:ns:meta/">
</x:xmpmeta>
"""


def test_xmp_no_rdf_in_xmpmeta():
    with pytest.raises(
        xmp_xml.XmpXmlProcessingError, match="RDF node in x:xmpmeta"
    ):
        xmp_xml.parse_xmp(BytesIO(XMP_NO_RDF_IN_XMPMETA.encode('utf8')))


XMP_INVALID_ROOT = """
<xml:boo/>
"""


def test_xmp_invalid_root():
    with pytest.raises(xmp_xml.XmpXmlProcessingError, match="XMP root must be"):
        xmp_xml.parse_xmp(BytesIO(XMP_INVALID_ROOT.encode('utf8')))


# noinspection HttpUrlsUsage
XMP_WITH_DUPLICATE_STRUCT_FIELD = """
<rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#"
         xmlns:xe="http://ns.adobe.com/xmp-example/">
    <rdf:Description rdf:about="">
        <xe:title>bleh</xe:title>
        <xe:title>bleh</xe:title>
   </rdf:Description>
</rdf:RDF>
"""


def test_xmp_with_duplicate_struct_field():
    with pytest.raises(xmp_xml.XmpXmlProcessingError, match="Duplicate field"):
        xmp_xml.parse_xmp(
            BytesIO(XMP_WITH_DUPLICATE_STRUCT_FIELD.encode('utf8'))
        )


# noinspection HttpUrlsUsage
XMP_INVALID_DATE1 = """
<x:xmpmeta xmlns:x="adobe:ns:meta/">
 <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#">
  <rdf:Description rdf:about="" xmlns:xmp="http://ns.adobe.com/xap/1.0/">
   <xmp:CreateDate>1/1/2022</xmp:CreateDate>
  </rdf:Description>
 </rdf:RDF>
</x:xmpmeta>
"""


def test_xmp_invalid_date1():
    with pytest.raises(
        xmp_xml.XmpXmlProcessingError, match="Failed to parse.*as a date"
    ):
        xmp_xml.meta_from_xmp(
            xmp_xml.parse_xmp(BytesIO(XMP_INVALID_DATE1.encode('utf8')))
        )


# noinspection HttpUrlsUsage
XMP_INVALID_DATE2 = """
<x:xmpmeta xmlns:x="adobe:ns:meta/">
 <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#">
  <rdf:Description rdf:about="" xmlns:xmp="http://ns.adobe.com/xap/1.0/">
   <xmp:CreateDate rdf:parseType="Resource"></xmp:CreateDate>
  </rdf:Description>
 </rdf:RDF>
</x:xmpmeta>
"""


def test_xmp_invalid_date2():
    with pytest.raises(xmp_xml.XmpXmlProcessingError, match="Wrong type"):
        xmp_xml.meta_from_xmp(
            xmp_xml.parse_xmp(BytesIO(XMP_INVALID_DATE2.encode('utf8')))
        )


# noinspection HttpUrlsUsage
PDFA_PLUS_UA_XMP_SAMPLE_WRONG_URI_TYPE = """
<x:xmpmeta xmlns:x="adobe:ns:meta/" x:xmptk="Adobe XMP Core 5.1.0-jc003">
  <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#">
    <rdf:Description rdf:about=""
        xmlns:dc="http://purl.org/dc/elements/1.1/"
        xmlns:xmp="http://ns.adobe.com/xap/1.0/"
        xmlns:pdf="http://ns.adobe.com/pdf/1.3/"
        xmlns:pdfuaid="http://www.aiim.org/pdfua/ns/id/"
        xmlns:pdfaid="http://www.aiim.org/pdfa/ns/id/"
        xmlns:pdfaExtension="http://www.aiim.org/pdfa/ns/extension/"
        xmlns:pdfaSchema="http://www.aiim.org/pdfa/ns/schema#"
        xmlns:pdfaProperty="http://www.aiim.org/pdfa/ns/property#"
      dc:format="application/pdf"
      xmp:CreateDate="2022-07-07T01:26:05+02:00"
      xmp:ModifyDate="2022-07-07T01:26:05+02:00"
      pdfaid:part="2"
      pdfaid:conformance="A">
      <dc:creator>
        <rdf:Seq>
          <rdf:li>Matthias Valvekens</rdf:li>
        </rdf:Seq>
      </dc:creator>
      <dc:title>
        <rdf:Alt>
          <rdf:li xml:lang="x-default">Curriculum Vitae</rdf:li>
        </rdf:Alt>
      </dc:title>
      <pdfuaid:part>1</pdfuaid:part>
      <pdfaExtension:schemas>
        <rdf:Bag>
          <rdf:li>
            <rdf:Description
              pdfaSchema:namespaceURI="http://www.aiim.org/pdfua/ns/id/"
              pdfaSchema:prefix="pdfuaid"
              pdfaSchema:schema="PDF/UA identification schema">
            <pdfaSchema:property>
              <rdf:Seq>
                <rdf:li
                  pdfaProperty:category="internal"
                  pdfaProperty:description="PDF/UA version identifier"
                  pdfaProperty:name="part"
                  pdfaProperty:valueType="Integer"/>
                <rdf:li
                  pdfaProperty:category="internal"
                  pdfaProperty:description="PDF/UA amendment identifier"
                  pdfaProperty:name="amd"
                  pdfaProperty:valueType="Text"/>
                <rdf:li
                  pdfaProperty:category="internal"
                  pdfaProperty:description="PDF/UA corrigenda identifier"
                  pdfaProperty:name="corr"
                  pdfaProperty:valueType="Text"/>
              </rdf:Seq>
            </pdfaSchema:property>
            </rdf:Description>
          </rdf:li>
        </rdf:Bag>
      </pdfaExtension:schemas>
    </rdf:Description>
  </rdf:RDF>
</x:xmpmeta>
"""


# noinspection HttpUrlsUsage
PDFA_PLUS_UA_XMP_SAMPLE_CORRECT_URI_TYPE = """
<x:xmpmeta xmlns:x="adobe:ns:meta/" x:xmptk="Adobe XMP Core 5.1.0-jc003">
  <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#">
    <rdf:Description rdf:about=""
        xmlns:dc="http://purl.org/dc/elements/1.1/"
        xmlns:xmp="http://ns.adobe.com/xap/1.0/"
        xmlns:pdf="http://ns.adobe.com/pdf/1.3/"
        xmlns:pdfuaid="http://www.aiim.org/pdfua/ns/id/"
        xmlns:pdfaid="http://www.aiim.org/pdfa/ns/id/"
        xmlns:pdfaExtension="http://www.aiim.org/pdfa/ns/extension/"
        xmlns:pdfaSchema="http://www.aiim.org/pdfa/ns/schema#"
        xmlns:pdfaProperty="http://www.aiim.org/pdfa/ns/property#"
      dc:format="application/pdf"
      xmp:CreateDate="2022-07-07T01:26:05+02:00"
      xmp:ModifyDate="2022-07-07T01:26:05+02:00"
      pdfaid:part="2"
      pdfaid:conformance="A">
      <dc:creator>
        <rdf:Seq>
          <rdf:li>Matthias Valvekens</rdf:li>
        </rdf:Seq>
      </dc:creator>
      <dc:title>
        <rdf:Alt>
          <rdf:li xml:lang="x-default">Curriculum Vitae</rdf:li>
        </rdf:Alt>
      </dc:title>
      <pdfuaid:part>1</pdfuaid:part>
      <pdfaExtension:schemas>
        <rdf:Bag>
          <rdf:li>
            <rdf:Description
              pdfaSchema:prefix="pdfuaid"
              pdfaSchema:schema="PDF/UA identification schema">
            <pdfaSchema:namespaceURI
              rdf:resource="http://www.aiim.org/pdfua/ns/id/"/>
            <pdfaSchema:property>
              <rdf:Seq>
                <rdf:li
                  pdfaProperty:category="internal"
                  pdfaProperty:description="PDF/UA version identifier"
                  pdfaProperty:name="part"
                  pdfaProperty:valueType="Integer"/>
                <rdf:li
                  pdfaProperty:category="internal"
                  pdfaProperty:description="PDF/UA amendment identifier"
                  pdfaProperty:name="amd"
                  pdfaProperty:valueType="Text"/>
                <rdf:li
                  pdfaProperty:category="internal"
                  pdfaProperty:description="PDF/UA corrigenda identifier"
                  pdfaProperty:name="corr"
                  pdfaProperty:valueType="Text"/>
              </rdf:Seq>
            </pdfaSchema:property>
            </rdf:Description>
          </rdf:li>
        </rdf:Bag>
      </pdfaExtension:schemas>
    </rdf:Description>
  </rdf:RDF>
</x:xmpmeta>
"""


@pytest.mark.parametrize('sample', ['ok', 'wrong uri type'])
def test_parse_pdfa_ext_schema(sample):
    xmp = (
        PDFA_PLUS_UA_XMP_SAMPLE_CORRECT_URI_TYPE
        if sample == 'ok'
        else PDFA_PLUS_UA_XMP_SAMPLE_WRONG_URI_TYPE
    )
    roots = xmp_xml.parse_xmp(BytesIO(xmp.encode('utf8')))
    root = roots[0]
    pdfa_ext = root[model.ExpandedName(model.NS['pdfaExtension'], 'schemas')]
    pdfua_ext_schema = pdfa_ext.value.entries[0]
    pfx = pdfua_ext_schema.value[
        model.ExpandedName(model.NS['pdfaSchema'], 'prefix')
    ].value
    assert pfx == "pdfuaid"
    property_seq = pdfua_ext_schema.value[
        model.ExpandedName(model.NS['pdfaSchema'], 'property')
    ].value
    desc_value = property_seq.entries[0].value[
        model.ExpandedName(model.NS['pdfaProperty'], 'description')
    ]
    assert desc_value.value == "PDF/UA version identifier"
    ns_uri = pdfua_ext_schema.value[
        model.ExpandedName(model.NS['pdfaSchema'], 'namespaceURI')
    ].value
    assert isinstance(ns_uri, model.XmpUri)
    assert ns_uri.value == model.NS['pdfuaid']


@pytest.mark.parametrize('sample', ['ok', 'wrong uri type'])
def test_reser_pdfa_ext_schema(sample):
    xmp = (
        PDFA_PLUS_UA_XMP_SAMPLE_CORRECT_URI_TYPE
        if sample == 'ok'
        else PDFA_PLUS_UA_XMP_SAMPLE_WRONG_URI_TYPE
    )
    roots = xmp_xml.parse_xmp(BytesIO(xmp.encode('utf8')))
    out = BytesIO()
    xmp_xml.serialise_xmp(roots, out)
    roots_redux = xmp_xml.parse_xmp(out)
    assert roots_redux[0] == roots[0]


XMP_WITH_EMPTY_STRING = """
<x:xmpmeta xmlns:x="adobe:ns:meta/" x:xmptk="Adobe XMP Core 5.1.0-jc003">
  <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#">
    <rdf:Description rdf:about=""
        xmlns:dc="http://purl.org/dc/elements/1.1/"
        dc:format="application/pdf">
      <dc:creator>
        <rdf:Seq>
          <rdf:li/>
        </rdf:Seq>
      </dc:creator>
    </rdf:Description>
  </rdf:RDF>
</x:xmpmeta>
"""


def test_empty_string():
    roots = xmp_xml.parse_xmp(BytesIO(XMP_WITH_EMPTY_STRING.encode('utf8')))
    assert roots[0][model.DC_CREATOR].value.entries[0].value == ""


def test_xmp_lang_explicit_xdefault():
    w = copy_into_new_writer(PdfFileReader(BytesIO(MINIMAL)))
    w.output_version = (2, 0)
    w.document_meta.subject = model.StringWithLanguage("Blah", "DEFAULT")
    out = BytesIO()
    w.write(out)

    r = PdfFileReader(out)
    assert (
        b"<rdf:li xml:lang=\"x-default\">Blah</rdf:li>"
        in r.root['/Metadata'].data
    )


def test_encrypt_skipping_managed_metadata():
    w = copy_into_new_writer(PdfFileReader(BytesIO(MINIMAL)))
    w.encrypt("secret", "secret", encrypt_metadata=False)
    w.document_meta.title = "Metadata is unencrypted"
    out = BytesIO()
    w.write(out)

    r = PdfFileReader(out)
    # do not decrypt anything (not that it would matter in this case)
    ref = r.root.raw_get("/Metadata", decrypt=EncryptedObjAccess.RAW)
    raw_md = r.get_object(ref, never_decrypt=True)
    assert raw_md['/DecodeParms']['/Name'] == '/Identity'
    assert b'Metadata is unencrypted' in raw_md.encoded_data

    r = PdfFileReader(out)
    r.decrypt("secret")
    assert r.document_meta_view.title.value == "Metadata is unencrypted"


def test_encrypt_with_managed_metadata():
    w = copy_into_new_writer(PdfFileReader(BytesIO(MINIMAL)))
    w.encrypt("secret", "secret")
    w.document_meta.title = "Metadata is encrypted"
    out = BytesIO()
    w.write(out)

    r = PdfFileReader(out)
    # do not decrypt at first
    ref = r.root.raw_get("/Metadata", decrypt=EncryptedObjAccess.RAW)
    raw_md = r.get_object(ref, never_decrypt=True)
    assert b'Metadata is encrypted' not in raw_md.encoded_data

    r = PdfFileReader(out)
    r.decrypt("secret")
    assert r.document_meta_view.title.value == "Metadata is encrypted"


@freeze_time(datetime(2022, 9, 7, tzinfo=tzlocal.get_localzone()))
def test_incremental_update_handle_binary_producer_string():
    # Check whether binary (nondecodable) producer strings are handled
    # gracefully in incremental updates

    fname = os.path.join(PDF_DATA_DIR, 'info-bin-producer-string.pdf')
    out = BytesIO()
    with open(fname, 'rb') as inf:
        w = IncrementalPdfFileWriter(inf)
        w.root['/Blah'] = generic.pdf_name('/Blah')
        w.update_root()
        w.write(out)

    r = PdfFileReader(out)
    assert r.trailer_view['/Info']['/Producer'].startswith('pyHanko')


@freeze_time(datetime(2022, 9, 7, tzinfo=tzlocal.get_localzone()))
def test_rewrite_handle_binary_producer_string():
    # Check whether binary (nondecodable) producer strings are handled
    # gracefully

    fname = os.path.join(PDF_DATA_DIR, 'info-bin-producer-string.pdf')
    out = BytesIO()
    with open(fname, 'rb') as inf:
        w = copy_into_new_writer(PdfFileReader(inf))
        w.root['/Blah'] = generic.pdf_name('/Blah')
        w.write(out)

    r = PdfFileReader(out)
    assert r.trailer_view['/Info']['/Producer'].startswith('pyHanko')


def test_bogus_metadata_key_value():
    out = BytesIO(MINIMAL)
    w = IncrementalPdfFileWriter(out)
    w.root['/Metadata'] = generic.pdf_name('/Blah')
    w.update_root()
    w._update_meta = lambda: None
    w.write_in_place()

    r = PdfFileReader(out)
    assert r._xmp_meta_view() is None



from io import BytesIO

import pytest
from freezegun import freeze_time
from pyhanko.pdf_utils.incremental_writer import IncrementalPdfFileWriter
from pyhanko.pdf_utils.reader import PdfFileReader
from pyhanko.pdf_utils.writer import copy_into_new_writer
from pyhanko.sign import signers
from pyhanko.sign.diff_analysis import ModificationLevel
from pyhanko.sign.signers.pdf_signer import (
    DSSContentSettings,
    SigDSSPlacementPreference,
)
from pyhanko.sign.validation import validate_pdf_signature

from .samples import (
    MINIMAL_AES256,
    MINIMAL_ONE_FIELD_AES256,
    MINIMAL_ONE_FIELD_RC4,
    MINIMAL_PUBKEY_ONE_FIELD_AES256,
    MINIMAL_PUBKEY_ONE_FIELD_RC4,
    MINIMAL_RC4,
    PDF_DATA_DIR,
    PUBKEY_SELFSIGNED_DECRYPTER,
)
from .signing_commons import (
    DUMMY_HTTP_TS,
    FROM_CA,
    SIMPLE_V_CONTEXT,
    live_testing_vc,
    val_trusted,
)
from .test_pades import PADES

sign_crypt_rc4_files = (MINIMAL_RC4, MINIMAL_ONE_FIELD_RC4)
sign_crypt_aes256_files = (MINIMAL_AES256, MINIMAL_ONE_FIELD_AES256)


@pytest.mark.parametrize('password', [b'usersecret', b'ownersecret'])
@freeze_time('2020-11-01')
def test_sign_crypt_rc4(password):
    w = IncrementalPdfFileWriter(BytesIO(MINIMAL_ONE_FIELD_RC4))
    w.encrypt(password)
    out = signers.sign_pdf(
        w,
        signers.PdfSignatureMetadata(),
        signer=FROM_CA,
        existing_fields_only=True,
    )

    r = PdfFileReader(out)
    r.decrypt(password)
    s = r.embedded_signatures[0]
    val_trusted(s)


@pytest.mark.parametrize('password', ['usersecret', 'ownersecret'])
@freeze_time('2020-11-01')
def test_sign_crypt_aes256(password):
    w = IncrementalPdfFileWriter(BytesIO(MINIMAL_ONE_FIELD_AES256))
    w.encrypt(password)
    out = signers.sign_pdf(
        w,
        signers.PdfSignatureMetadata(),
        signer=FROM_CA,
        existing_fields_only=True,
    )

    r = PdfFileReader(out)
    r.decrypt(password)
    s = r.embedded_signatures[0]
    val_trusted(s)


@freeze_time('2020-11-01')
def test_sign_crypt_pubkey_aes256():
    w = IncrementalPdfFileWriter(BytesIO(MINIMAL_PUBKEY_ONE_FIELD_AES256))
    w.encrypt_pubkey(PUBKEY_SELFSIGNED_DECRYPTER)
    out = signers.sign_pdf(
        w,
        signers.PdfSignatureMetadata(),
        signer=FROM_CA,
        existing_fields_only=True,
    )

    r = PdfFileReader(out)
    r.decrypt_pubkey(PUBKEY_SELFSIGNED_DECRYPTER)
    s = r.embedded_signatures[0]
    val_trusted(s)


@freeze_time('2020-11-01')
def test_sign_crypt_pubkey_rc4():
    w = IncrementalPdfFileWriter(BytesIO(MINIMAL_PUBKEY_ONE_FIELD_RC4))
    w.encrypt_pubkey(PUBKEY_SELFSIGNED_DECRYPTER)
    out = signers.sign_pdf(
        w,
        signers.PdfSignatureMetadata(),
        signer=FROM_CA,
        existing_fields_only=True,
    )

    r = PdfFileReader(out)
    r.decrypt_pubkey(PUBKEY_SELFSIGNED_DECRYPTER)
    s = r.embedded_signatures[0]
    val_trusted(s)


sign_crypt_new_params = [
    [b'usersecret', 0],
    [b'usersecret', 1],
    [b'ownersecret', 0],
    [b'ownersecret', 1],
]


@pytest.mark.parametrize('password, file', sign_crypt_new_params)
@freeze_time('2020-11-01')
def test_sign_crypt_rc4_new(password, file):
    w = IncrementalPdfFileWriter(BytesIO(sign_crypt_rc4_files[file]))
    w.encrypt(password)
    out = signers.sign_pdf(
        w,
        signers.PdfSignatureMetadata(field_name='SigNew'),
        signer=FROM_CA,
    )
    out.seek(0)
    r = PdfFileReader(out)
    r.decrypt(password)

    s = r.embedded_signatures[0]
    val_trusted(s)


@pytest.mark.parametrize('password, file', sign_crypt_new_params)
@freeze_time('2020-11-01')
def test_sign_crypt_aes256_new(password, file):
    w = IncrementalPdfFileWriter(BytesIO(sign_crypt_aes256_files[file]))
    w.encrypt(password)
    out = signers.sign_pdf(
        w,
        signers.PdfSignatureMetadata(field_name='SigNew'),
        signer=FROM_CA,
    )
    out.seek(0)
    r = PdfFileReader(out)
    r.decrypt(password)

    s = r.embedded_signatures[0]
    val_trusted(s)


@pytest.mark.parametrize('password, file', sign_crypt_new_params)
@freeze_time('2020-11-01')
def test_sign_encrypted_with_post_sign(requests_mock, password, file):
    w = IncrementalPdfFileWriter(BytesIO(sign_crypt_aes256_files[file]))
    w.encrypt(password)

    out = signers.sign_pdf(
        w,
        signers.PdfSignatureMetadata(
            field_name='Sig1',
            validation_context=live_testing_vc(requests_mock),
            subfilter=PADES,
            embed_validation_info=True,
            dss_settings=DSSContentSettings(
                placement=SigDSSPlacementPreference.SEPARATE_REVISION
            ),
            use_pades_lta=True,
        ),
        signer=FROM_CA,
        timestamper=DUMMY_HTTP_TS,
    )
    r = PdfFileReader(out)
    r.decrypt(password)

    s = r.embedded_signatures[0]
    status = val_trusted(s, extd=True)
    assert status.modification_level == ModificationLevel.LTA_UPDATES
    assert len(r.embedded_regular_signatures) == 1
    assert len(r.embedded_timestamp_signatures) == 1


def test_copy_encrypted_signed_file():
    w = IncrementalPdfFileWriter(BytesIO(MINIMAL_ONE_FIELD_AES256))
    w.encrypt("ownersecret")
    out = signers.sign_pdf(
        w,
        signers.PdfSignatureMetadata(),
        signer=FROM_CA,
        existing_fields_only=True,
    )

    r = PdfFileReader(out)
    r.decrypt("ownersecret")
    w = copy_into_new_writer(r)
    out2 = BytesIO()
    w.write(out2)

    r = PdfFileReader(out2)
    assert not r.encrypted
    s = r.embedded_signatures[0]
    s.compute_integrity_info()
    status = validate_pdf_signature(s, SIMPLE_V_CONTEXT(), skip_diff=True)
    assert not status.intact


def test_copy_file_with_mdp_signature_and_backref():
    # This file has /Data in a signature reference dictionary
    # pointing back to the root (which is sometimes still seen in
    # FieldMDP signatures generated by Acrobat, among others)

    fname = f"{PDF_DATA_DIR}/signed-encrypted-pubkey-with-catalog-ref.pdf"
    with open(fname, 'rb') as inf:

        r = PdfFileReader(inf)
        r.decrypt_pubkey(PUBKEY_SELFSIGNED_DECRYPTER)

        w = copy_into_new_writer(r)
        out2 = BytesIO()
        w.write(out2)

    r = PdfFileReader(out2)
    assert not r.encrypted
    s = r.embedded_signatures[0]
    s.compute_integrity_info()
    status = validate_pdf_signature(s, SIMPLE_V_CONTEXT(), skip_diff=True)
    assert not status.intact



from typing import Optional, Type

from asn1crypto.algos import DigestAlgorithmId
from asn1crypto.core import ObjectIdentifier

from ..ades.report import AdESIndeterminate, AdESStatus, AdESSubIndic
from ..general import ValueErrorWithMessage

__all__ = [
    'SignatureValidationError',
    'DisallowedAlgorithmError',
    'ValidationInfoReadingError',
    'NoDSSFoundError',
    'SigSeedValueValidationError',
    'CMSAlgorithmProtectionError',
]


class ValidationInfoReadingError(ValueErrorWithMessage):
    """Error reading validation info."""

    pass


class NoDSSFoundError(ValidationInfoReadingError):
    def __init__(self):
        super().__init__("No DSS found")


class CMSAlgorithmProtectionError(ValueErrorWithMessage):
    """Error related to CMS algorithm protection checks."""


class SignatureValidationError(ValueErrorWithMessage):
    """Error validating a signature."""

    def __init__(
        self, failure_message, ades_subindication: Optional[AdESSubIndic] = None
    ):
        self.ades_subindication = ades_subindication
        if ades_subindication:
            msg = "%s [%s]" % (failure_message, ades_subindication)
        else:
            msg = failure_message
        super().__init__(msg)

    @property
    def ades_status(self) -> Optional[AdESStatus]:
        if self.ades_subindication is not None:
            return self.ades_subindication.status
        return None


class DisallowedAlgorithmError(SignatureValidationError):
    def __init__(
        self,
        failure_message,
        permanent: bool,
        oid_type: Optional[Type[ObjectIdentifier]] = None,
    ):
        self.oid_type = oid_type
        if permanent:
            subindic = AdESIndeterminate.CRYPTO_CONSTRAINTS_FAILURE
        else:
            subindic = AdESIndeterminate.CRYPTO_CONSTRAINTS_FAILURE_NO_POE
        super().__init__(
            failure_message=failure_message, ades_subindication=subindic
        )


class SigSeedValueValidationError(SignatureValidationError):
    """Error validating a signature's seed value constraints."""

    # TODO perhaps we can encode some more metadata here, such as the
    #  seed value that tripped the failure.
    pass



import getpass
from typing import BinaryIO, Optional

import click
from pyhanko.pdf_utils import crypt
from pyhanko.pdf_utils.crypt import AuthStatus
from pyhanko.pdf_utils.incremental_writer import IncrementalPdfFileWriter


class OpenForSigning:
    def __init__(self, infile_path: str, lenient: bool):
        self.infile_path = infile_path
        self.lenient = lenient
        self.handle: Optional[BinaryIO] = None

    def __enter__(self) -> IncrementalPdfFileWriter:
        self.handle = infile = open(self.infile_path, 'rb')
        writer = IncrementalPdfFileWriter(infile, strict=not self.lenient)

        # TODO make this an option higher up the tree
        if writer.prev.encrypted:
            sh = writer.prev.security_handler
            if isinstance(sh, crypt.StandardSecurityHandler):
                pdf_pass = getpass.getpass(
                    prompt='Password for encrypted file \'%s\': '
                    % self.infile_path
                )
                auth = writer.encrypt(pdf_pass)
                if auth.status == AuthStatus.FAILED:
                    raise click.ClickException(
                        "Invalid password for encrypted file"
                    )
            elif isinstance(sh, crypt.PubKeySecurityHandler):
                raise click.ClickException(
                    "Public-key document encryption is not supported in the CLI"
                )
            else:  # pragma: nocover
                raise click.ClickException(
                    "Input file appears to be encrypted, but appropriate "
                    "credentials are not available."
                )
        return writer

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self.handle:
            self.handle.close()


def open_for_signing(infile_path: str, lenient: bool):
    return OpenForSigning(infile_path, lenient)


def get_text_params(ctx):
    text_params = None
    stamp_url = ctx.obj.stamp_url
    if stamp_url is not None:
        text_params = {'url': stamp_url}
    return text_params



"""
Module for AdES reporting data.

Defines enums for all AdES validation statuses defined in ETSI EN 319 102-1,
clause 5.1.3.
"""

import enum

__all__ = [
    'AdESStatus',
    'AdESSubIndic',
    'AdESPassed',
    'AdESFailure',
    'AdESIndeterminate',
]


# TODO document these


class AdESStatus(enum.Enum):
    PASSED = enum.auto()
    INDETERMINATE = enum.auto()
    FAILED = enum.auto()


class AdESSubIndic:
    @property
    def status(self) -> AdESStatus:
        raise NotImplementedError

    @property
    def standard_name(self):
        raise NotImplementedError


class AdESPassed(AdESSubIndic, enum.Enum):
    OK = enum.auto()

    @property
    def status(self) -> AdESStatus:
        return AdESStatus.PASSED

    @property
    def standard_name(self):
        return self.name


class AdESFailure(AdESSubIndic, enum.Enum):
    FORMAT_FAILURE = enum.auto()
    HASH_FAILURE = enum.auto()
    SIG_CRYPTO_FAILURE = enum.auto()
    REVOKED = enum.auto()
    NOT_YET_VALID = enum.auto()

    @property
    def status(self):
        return AdESStatus.FAILED

    @property
    def standard_name(self):
        return self.name


class AdESIndeterminate(AdESSubIndic, enum.Enum):
    SIG_CONSTRAINTS_FAILURE = enum.auto()
    CHAIN_CONSTRAINTS_FAILURE = enum.auto()
    CERTIFICATE_CHAIN_GENERAL_FAILURE = enum.auto()
    CRYPTO_CONSTRAINTS_FAILURE = enum.auto()
    EXPIRED = enum.auto()
    NOT_YET_VALID = enum.auto()
    POLICY_PROCESSING_ERROR = enum.auto()
    SIGNATURE_POLICY_NOT_AVAILABLE = enum.auto()
    TIMESTAMP_ORDER_FAILURE = enum.auto()
    NO_SIGNING_CERTIFICATE_FOUND = enum.auto()
    NO_CERTIFICATE_CHAIN_FOUND = enum.auto()
    REVOKED_NO_POE = enum.auto()
    REVOKED_CA_NO_POE = enum.auto()
    OUT_OF_BOUNDS_NO_POE = enum.auto()
    REVOCATION_OUT_OF_BOUNDS_NO_POE = enum.auto()
    OUT_OF_BOUNDS_NOT_REVOKED = enum.auto()
    CRYPTO_CONSTRAINTS_FAILURE_NO_POE = enum.auto()
    NO_POE = enum.auto()
    TRY_LATER = enum.auto()
    SIGNED_DATA_NOT_FOUND = enum.auto()
    GENERIC = enum.auto()

    @property
    def status(self):
        return AdESStatus.INDETERMINATE

    @property
    def standard_name(self):
        return self.name



import itertools
import math
from typing import List, Optional

import qrcode.util
from pyhanko.pdf_utils.content import PdfContent
from pyhanko.pdf_utils.misc import rd
from qrcode.image.base import BaseImage


class PdfStreamQRImage(BaseImage):
    """
    Quick-and-dirty implementation of the Image interface required
    by the qrcode package.
    """

    kind = "PDF"
    allowed_kinds = ("PDF",)
    qr_color = (0, 0, 0)

    def new_image(self, **kwargs):
        return []

    def drawrect(self, row, col):
        self._img.append((row, col))

    def append_single_rect(self, command_stream, row, col):
        command_stream.append(b'%g %g 1 1 re' % (col, row))

    def format_qr_color(self):
        return (b"%g %g %g rg\n" % self.qr_color) + (
            b"%g %g %g RG" % self.qr_color
        )

    def setup_drawing_area(self):
        # start a command stream with fill colour set to black (default)
        # and transform the coordinate system to line up with our grid
        brd = rd(self.border * self.box_size)
        ydiff = rd(self.width * self.box_size)
        cm = f"{rd(self.box_size)} 0 0 {-rd(self.box_size)} {brd} {brd + ydiff} cm"
        return b"%s\n%s" % (self.format_qr_color(), cm.encode('ascii'))

    def render_command_stream(self):
        command_stream = [self.setup_drawing_area()]
        for row, col in self._img:
            # paint a rectangle
            self.append_single_rect(command_stream, row, col)
        command_stream.append(b"f")
        return b'\n'.join(command_stream)

    def save(self, stream, kind=None):
        raise NotImplementedError

    def process(self):
        raise NotImplementedError

    def drawrect_context(self, row, col, active, context):
        return self.drawrect(row, col)  # pragma: nocover


class PdfFancyQRImage(PdfStreamQRImage):
    centerpiece_corner_radius = 0.2

    def __init__(
        self,
        border,
        width,
        box_size,
        *_args,
        version,
        center_image: Optional[PdfContent] = None,
        **kwargs,
    ):
        super().__init__(border, width, box_size, **kwargs)
        self._version = version
        self._centerpiece = center_image

    def save(self, stream, kind=None):
        raise NotImplementedError

    def process(self):
        raise NotImplementedError

    def append_single_rect(self, command_stream, row, col):
        if self.is_position_pattern(row, col):
            return
        command_stream.extend(rounded_square(col, row, 0.9, 0.3))

    def is_major_position_pattern(self, row, col):
        return (
            (row < 7 and col < 7)
            or ((row > self.width - 8) and col < 7)
            or (row < 7 and (col > self.width - 8))
        )

    def _enumerate_alignment_patterns(self):
        adj_ptns = qrcode.util.pattern_position(self._version)
        for pr, pc in itertools.product(adj_ptns, adj_ptns):
            # don't consider the bits that fall within the "big" position
            # patterns
            if self.is_major_position_pattern(pr, pc):
                continue
            yield pr, pc

    def is_position_pattern(self, row, col):
        if self.is_major_position_pattern(row, col):
            return True
        # check whether this pixel is part of an alignment pattern
        return any(
            abs(row - pr) <= 2 and abs(col - pc) <= 2
            for pr, pc in self._enumerate_alignment_patterns()
        )

    def draw_position_patterns(self):
        # draw the surrounding squares of the location patterns first
        #  -> stroked paths (outer squares)
        # (these require drawing at module midpoints)
        command_stream = [b'q\n1 0 0 1 0.5 0.5 cm\n0.7 w']
        sz = self.width
        command_stream.extend(rounded_square(0, 0, 6, 1))
        command_stream.extend(rounded_square(0, sz - 7, 6, 1))
        command_stream.extend(rounded_square(sz - 7, 0, 6, 1))
        for pr, pc in self._enumerate_alignment_patterns():
            command_stream.extend(rounded_square(pr - 2, pc - 2, 4, 0.7))
        command_stream.append(b"S\nQ")

        # draw the inner squares of the location patterns
        command_stream.extend(rounded_square(2, 2, 3, 0.6))
        command_stream.extend(rounded_square(2, sz - 7 + 2, 3, 0.6))
        command_stream.extend(rounded_square(sz - 7 + 2, 2, 3, 0.6))

        for pr, pc in self._enumerate_alignment_patterns():
            command_stream.extend(rounded_square(pr, pc, 1, 0.1))
        command_stream.append(b"f")
        return b"\n".join(command_stream)

    def draw_centerpiece(self):
        c_x, c_y, c_sz = self._measure_out_centerpiece()
        # better hope it's square
        c_w = self._centerpiece.box.width
        c_h = self._centerpiece.box.height

        # draw the border of the centerpiece
        centerpiece_commands = [b"q", b"0.2 w"]
        centerpiece_commands.extend(
            rounded_square(
                c_x, c_y, c_sz, self.centerpiece_corner_radius * c_sz
            )
        )
        centerpiece_commands.append(b"S\nQ\nq")
        # transform back into the internal coordinates of the centerpiece
        # (including any y-axis reversals)
        # -> f"{x_scale} 0 0 {-y_scale} {c_x} {c_y + c_sz} cm"
        x_scale = rd(c_sz / c_w)
        y_scale = rd(c_sz / c_h)
        # COMBINED WITH:
        # we slightly shrink and offset the centerpiece to create
        # a border of sorts
        # -> f"{shrink} 0 0 {shrink} {x_shift} {y_shift} cm"
        shrink = 0.85
        x_shift = rd((1 - shrink) * c_w / 2)
        y_shift = rd((1 - shrink) * c_h / 2)
        centerpiece_commands.append(
            f"{x_scale * shrink} 0 0 {-y_scale * shrink} "
            f"{c_x + x_shift * x_scale} {c_y + c_sz - y_shift * y_scale} "
            f"cm".encode('ascii')
        )

        # resource management left up to the caller
        centerpiece_commands.append(self._centerpiece.render())
        centerpiece_commands.append(b"Q")
        return b"\n".join(centerpiece_commands)

    def _measure_out_centerpiece(self):
        # Centerpiece area takes up a square in the QR code with a side of
        # about 28% the total size of the QR code itself
        c_sz = 0.28 * self.width
        c_x = (self.width - c_sz) / 2
        c_y = (self.width - c_sz) / 2
        return rd(c_x), rd(c_y), rd(c_sz)

    def setup_drawing_area(self):
        basic_setup = super().setup_drawing_area()
        commands = [basic_setup]
        if self._centerpiece is not None:
            # we need to clip out the centerpiece area.
            # we'll do that by creating a clip path that goes around the
            # entire QR code, and a rounded rectangle where the centerpiece
            # would go using opposite orientation.

            # -> clockwise rectangle
            w = rd(self.width)
            # save the state to undo the clipping later
            commands.append(b"q")
            commands.append(b"0.2 w")
            commands.append(
                f"0 0 m 0 {w} l {w} {w} l {w} 0 l h".encode("ascii"),
            )
            c_x, c_y, c_sz = self._measure_out_centerpiece()
            commands.extend(
                rounded_square(
                    c_x, c_y, c_sz, self.centerpiece_corner_radius * c_sz
                )
            )
            commands.append(b"W n")
        return b"\n".join(commands)

    def render_command_stream(self):
        parts = [super().render_command_stream(), self.draw_position_patterns()]
        if self._centerpiece:
            parts.append(b"Q")  # undo clipping
            parts.append(self.draw_centerpiece())

        return b"\n".join(parts)


def rounded_square(
    x_pos: float, y_pos: float, sz: float, rad: float
) -> List[bytes]:
    """
    Add a subpath of a square with rounded corners at the given position.
    Doesn't include any painting or clipping operations.

    The path is drawn counterclockwise.

    :param x_pos:
        The x-coordinate of the enveloping square's lower left corner.
    :param y_pos:
        The y-coordinate of the enveloping square's lower left corner.
    :param sz:
        The side length of the enveloping square.
    :param rad:
        The corner radius.

    :return:
        A list of graphics operators.
    """

    c_off = (4 * (math.sqrt(2) - 1) / 3) * rad

    result = []

    def fmt(x, y):
        px = rd(x_pos + x)
        py = rd(y_pos + y)
        result.append(f"{px} {py} ".encode('ascii'))

    def op(pts, opc: str):
        for x, y in pts:
            fmt(x, y)
        result.append(opc.encode('ascii'))

    def uop(x, y, opc):
        op([(x, y)], opc)

    uop(rad, 0, "m")
    uop(sz - rad, 0, "l")
    op([(sz - c_off, 0), (sz, c_off), (sz, rad)], "c")
    uop(sz, sz - rad, "l")
    op([(sz, sz - c_off), (sz - c_off, sz), (sz - rad, sz)], "c")
    uop(rad, sz, "l")
    op([(c_off, sz), (0, sz - c_off), (0, sz - rad)], "c")
    uop(0, rad, "l")
    op([(0, c_off), (c_off, 0), (rad, 0)], "c")
    result.append(b"h")
    return result
